# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception


import ast
import re

import pytest

from tapl_lang.core import syntax
from tapl_lang.core.parser import Grammar, parse_text
from tapl_lang.lib import compiler, kinds, python_backend, scope, tapl_typing, terms
from tapl_lang.pythonlike import grammar, predef1, rule_names


def check_parsed_term(parsed: syntax.Term) -> None:
    if parsed is None:
        raise RuntimeError('Parser returns None.')
    error_bucket: list[syntax.ErrorTerm] = compiler.gather_errors(parsed)
    if error_bucket:
        messages = [e.message for e in error_bucket]
        raise SyntaxError('\n\n'.join(messages))


def parse_expr(text: str, *, debug=False) -> list[ast.expr]:
    parsed = parse_text(text, Grammar(grammar.get_grammar().rule_map, rule_names.EXPRESSION), debug=debug)
    check_parsed_term(parsed)
    safe_term = compiler.make_safe_term(parsed)
    separated = syntax.LayerSeparator(2).build(lambda layer: layer(safe_term))
    return [
        python_backend.AstGenerator().generate_expr(layer, syntax.BackendSetting(scope_level=0)) for layer in separated
    ]


def evaluate(expr: ast.expr, locals_=None):
    compiled_code = compile(ast.Expression(body=expr), filename='', mode='eval')
    return eval(compiled_code, {}, locals_ or {})


def typecheck(expr: ast.expr, locals_=None):
    compiled_code = compile(ast.Expression(body=expr), filename='', mode='eval')
    scope0 = scope.Scope(parent=predef1.predef_scope)
    scope0.store_many__sa(locals_ or {})
    globals_ = {'s0': scope0, 'tapl_typing': tapl_typing}
    return eval(compiled_code, globals_)


def expect_type_error(expr: ast.expr) -> str:
    try:
        typecheck(expr)
    except TypeError as e:
        return repr(e)
    else:
        return 'NO ERROR'


def test_constant_true():
    [expr1, expr2] = parse_expr('True')
    assert ast.unparse(expr1) == 'True'
    assert ast.unparse(expr2) == 's0.Bool'
    assert typecheck(expr2) is predef1.predef_scope.Bool
    assert evaluate(expr1) is True


def test_constant_number7():
    [expr1, expr2] = parse_expr('7')
    assert ast.unparse(expr1) == '7'
    assert ast.unparse(expr2) == 's0.Int'
    assert typecheck(expr2) is predef1.predef_scope.Int
    assert evaluate(expr1) == 7


def test_inversion_bool():
    [expr1, expr2] = parse_expr('not True')
    assert ast.unparse(expr1) == 'not True'
    assert ast.unparse(expr2) == 's0.Bool'
    assert typecheck(expr2) is predef1.predef_scope.Bool
    assert evaluate(expr1) is False


def test_inversion_number():
    [expr1, expr2] = parse_expr('not 0')
    assert ast.unparse(expr1) == 'not 0'
    assert ast.unparse(expr2) == 's0.Bool'
    assert typecheck(expr2) is predef1.predef_scope.Bool
    assert evaluate(expr1) is True


def test_negation_int():
    [expr1, expr2] = parse_expr('-7')
    assert ast.unparse(expr1) == '-7'
    assert ast.unparse(expr2) == '-s0.Int'
    assert typecheck(expr2) is predef1.predef_scope.Int
    assert evaluate(expr1) == -7


def test_negation_float():
    [expr1, expr2] = parse_expr('-7.3')
    assert ast.unparse(expr1) == '-7.3'
    assert ast.unparse(expr2) == '-s0.Float'
    assert typecheck(expr2) is predef1.predef_scope.Float
    assert evaluate(expr1) == -7.3


def test_conjuction_bool1():
    [expr1, expr2] = parse_expr('True and True')
    assert ast.unparse(expr1) == 'True and True'
    assert ast.unparse(expr2) == 'tapl_typing.create_union(s0.Bool, s0.Bool)'
    assert typecheck(expr2) is predef1.predef_scope.Bool
    assert evaluate(expr1) is True


def test_conjuction_bool2():
    [expr1, expr2] = parse_expr('True and True     and    False')
    assert ast.unparse(expr1) == 'True and True and False'
    assert ast.unparse(expr2) == 'tapl_typing.create_union(s0.Bool, s0.Bool, s0.Bool)'
    assert typecheck(expr2) is predef1.predef_scope.Bool
    assert evaluate(expr1) is False


def test_conjuction_number1():
    [expr1, expr2] = parse_expr('3 and 4')
    assert ast.unparse(expr1) == '3 and 4'
    assert ast.unparse(expr2) == 'tapl_typing.create_union(s0.Int, s0.Int)'
    assert typecheck(expr2) is predef1.predef_scope.Int
    assert evaluate(expr1) == 4


def test_conjuction_number2():
    [expr1, expr2] = parse_expr('0 and 4')
    assert ast.unparse(expr1) == '0 and 4'
    assert ast.unparse(expr2) == 'tapl_typing.create_union(s0.Int, s0.Int)'
    assert typecheck(expr2) is predef1.predef_scope.Int
    assert evaluate(expr1) == 0


def test_conjuction_mix():
    [expr1, expr2] = parse_expr('True and 4')
    assert ast.unparse(expr1) == 'True and 4'
    assert ast.unparse(expr2) == 'tapl_typing.create_union(s0.Bool, s0.Int)'
    assert kinds.check_type_equality(
        typecheck(expr2),
        tapl_typing.create_union(predef1.predef_scope.Int, predef1.predef_scope.Bool),
    )
    assert evaluate(expr1) == 4


def test_disjunction():
    [expr1, expr2] = parse_expr('True and True     and    False  or True')
    assert ast.unparse(expr1) == 'True and True and False or True'
    assert (
        ast.unparse(expr2) == 'tapl_typing.create_union(tapl_typing.create_union(s0.Bool, s0.Bool, s0.Bool), s0.Bool)'
    )
    assert kinds.check_type_equality(typecheck(expr2), predef1.predef_scope.Bool)
    assert evaluate(expr1) is True


def test_term1():
    [expr1, expr2] = parse_expr('2 + 3')
    assert ast.unparse(expr1) == '2 + 3'
    assert ast.unparse(expr2) == 's0.Int + s0.Int'
    assert kinds.check_type_equality(typecheck(expr2), predef1.predef_scope.Int)
    assert evaluate(expr1) == 5


def test_term2():
    [expr1, expr2] = parse_expr('2 + True')
    assert ast.unparse(expr1) == '2 + True'
    assert ast.unparse(expr2) == 's0.Int + s0.Bool'
    with pytest.raises(TypeError, match=r'unsupported operand type\(s\) for \+: Int and Bool'):
        typecheck(expr2)


def test_term_error():
    [expr1, expr2] = parse_expr("2 + 'abc'")
    assert ast.unparse(expr1) == "2 + 'abc'"
    assert ast.unparse(expr2) == 's0.Int + s0.Str'
    assert (
        expect_type_error(expr2)
        == "TypeError('unsupported operand type(s) for +: Int and Str. Function positional arguments are not equal: expected=[Int] actual=[Str]')"
    )


def test_compare1():
    [expr1, expr2] = parse_expr('2 < 3')
    assert ast.unparse(expr1) == '2 < 3'
    assert ast.unparse(expr2) == 's0.Int < s0.Int'
    assert kinds.check_type_equality(typecheck(expr2), predef1.predef_scope.Bool)
    assert evaluate(expr1) is True


def test_compare2():
    [expr1, expr2] = parse_expr('True < 0')
    assert ast.unparse(expr1) == 'True < 0'
    assert ast.unparse(expr2) == 's0.Bool < s0.Int'
    with pytest.raises(
        TypeError,
        match=re.escape(
            'unsupported operand type(s) for <: Bool and Int. Function positional arguments are not equal: expected=[Bool] actual=[Int]'
        ),
    ):
        typecheck(expr2)


def test_var1():
    [expr1, expr2] = parse_expr('2 + x')
    assert ast.unparse(expr1) == '2 + x'
    assert ast.unparse(expr2) == 's0.Int + s0.x'
    assert typecheck(expr2, locals_={'x': predef1.predef_scope.Int}) == predef1.predef_scope.Int
    assert evaluate(expr1, locals_={'x': 7}) == 9


def test_term_repr():
    parsed = parse_text('2+x', Grammar(grammar.get_grammar().rule_map, rule_names.EXPRESSION))
    assert (
        str(parsed)
        == "BinOp(left=IntegerLiteral(value=2, mode=Layers(layers=[ModeTerm(typecheck=False, use_scope=False), ModeTerm(typecheck=True, use_scope=True)]), location=(1:0,1:1)), op='+', right=TypedName(id='x', ctx='load', mode=Layers(layers=[ModeTerm(typecheck=False, use_scope=False), ModeTerm(typecheck=True, use_scope=True)]), location=(1:2,1:3)), location=(1:0,1:3))"
    )


def test_multi_layer_call():
    [expr1, expr2] = parse_expr('f(<3:Str>)')
    assert ast.unparse(expr1) == 'f(3)'
    assert ast.unparse(expr2) == 's0.f(s0.Str)'


def test_gather_errors():
    location = syntax.Location(start=syntax.Position(line=1, column=0))
    a = terms.IntegerLiteral(value=2, mode=terms.MODE_SAFE, location=location)
    b = terms.IntegerLiteral(value=3, mode=terms.MODE_SAFE, location=location)
    c = syntax.ErrorTerm('Expected number')
    d = terms.BinOp(b, '*', c, location=location)
    e = terms.BinOp(a, '+', d, location=location)
    error_bucket = compiler.gather_errors(e)
    assert len(error_bucket) == 1
    assert isinstance(error_bucket[0], syntax.ErrorTerm)
    assert error_bucket[0].message == 'Expected number'



from tapl_lang.lib import tapl_typing
from tapl_lang.pythonlike.predef1 import predef_scope as predef_scope__sa
s0 = tapl_typing.create_scope(parent__sa=predef_scope__sa)

def double(i):
    s1 = tapl_typing.create_scope(parent__sa=s0, i=i)
    s1.double = tapl_typing.create_function([s1.Int], s1.Int)
    tapl_typing.set_return_type(s1, s1.Int)
    tapl_typing.add_return_type(s1, s1.i * s1.Int)
    return tapl_typing.get_return_type(s1)
s0.double = tapl_typing.create_function([s0.Int], double(s0.Int))

def square(i):
    s1 = tapl_typing.create_scope(parent__sa=s0, i=i)
    s1.square = tapl_typing.create_function([s1.Int], s1.Int)
    tapl_typing.set_return_type(s1, s1.Int)
    tapl_typing.add_return_type(s1, s1.i * s1.i)
    return tapl_typing.get_return_type(s1)
s0.square = tapl_typing.create_function([s0.Int], square(s0.Int))
s0.print(s0.square(s0.double(s0.Int)))
s0.print(s0.double(s0.square(s0.Int)))


from tapl_lang.lib import tapl_typing
from tapl_lang.pythonlike.predef1 import predef_scope as predef_scope__sa
s0 = tapl_typing.create_scope(parent__sa=predef_scope__sa)
with tapl_typing.scope_forker(s0) as f0:
    s1 = tapl_typing.fork_scope(f0)
    s1.result = s1.Int / s1.Int
    s1 = tapl_typing.fork_scope(f0)
    s1.print(s1.Str)
    s1 = tapl_typing.fork_scope(f0)
    s1.print(s1.Str)



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import pytest

from tapl_lang.lib import builtin_types, kinds

_any = builtin_types.Any
_bool = builtin_types.Bool
_int = builtin_types.Int
_float = builtin_types.Float


def test_bool_subtype_of_itself():
    assert kinds.check_subtype(_bool, _bool)


def test_bool_subtype_of_any():
    assert kinds.check_subtype(_bool, _any)


def test_any_subtype_of_bool():
    assert not kinds.check_subtype(_any, _bool)


def union_len(union: kinds.Union) -> int:
    return len(list(union.iter_types__sa()))


def test_create_union():
    # Multiple types
    union = kinds.create_union(_bool, _int)
    union_types = list(union.iter_types__sa())
    assert len(union_types) == 2
    assert union_types[0] is _bool
    assert union_types[1] is _int
    # Flatten union
    type_list = [_bool, kinds.Union(types=[_int, _float])]
    assert union_len(kinds.Union(types=type_list)) == 2
    assert union_len(kinds.create_union(*type_list)) == 3
    # Trim union
    type_list = [_bool, _int, _bool]
    assert union_len(kinds.Union(types=type_list)) == 3
    assert union_len(kinds.create_union(*type_list)) == 2


def test_function_parameters_invariants():
    with pytest.raises(TypeError, match=r'Function posonlyargs must be a list\.'):
        kinds.Function(posonlyargs=_any, args=[], result=_bool)
    with pytest.raises(ValueError, match=r'Function args must be a list of \(name, type\) pairs\.'):
        kinds.Function(
            posonlyargs=[],
            args=[('x', _any), _bool],
            result=_bool,
        )


def test_lazy_function_result():
    func = kinds.Function(posonlyargs=[_int], args=[], lazy_result=lambda: _bool)
    assert func.result__sa is _bool



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception


import ast

from tapl_lang.core import syntax, tapl_error
from tapl_lang.lib import terms

# Unary 'not' has a dedicated 'BoolNot' term for logical negation
UNARY_OP_MAP: dict[str, ast.unaryop] = {'+': ast.UAdd(), '-': ast.USub(), 'not': ast.Not(), '~': ast.Invert()}
BIN_OP_MAP: dict[str, ast.operator] = {
    '+': ast.Add(),
    '-': ast.Sub(),
    '*': ast.Mult(),
    '/': ast.Div(),
    '//': ast.FloorDiv(),
    '%': ast.Mod(),
    '**': ast.Pow(),
    '<<': ast.LShift(),
    '>>': ast.RShift(),
    '|': ast.BitOr(),
    '^': ast.BitXor(),
    '&': ast.BitAnd(),
    '@': ast.MatMult(),
}
BOOL_OP_MAP: dict[str, ast.boolop] = {'and': ast.And(), 'or': ast.Or()}
COMPARE_OP_MAP: dict[str, ast.cmpop] = {
    '==': ast.Eq(),
    '!=': ast.NotEq(),
    '<': ast.Lt(),
    '<=': ast.LtE(),
    '>': ast.Gt(),
    '>=': ast.GtE(),
    'is': ast.Is(),
    'is not': ast.IsNot(),
    'in': ast.In(),
    'not in': ast.NotIn(),
}
EXPR_CONTEXT_MAP: dict[str, ast.expr_context] = {'load': ast.Load(), 'store': ast.Store(), 'delete': ast.Del()}


def locate(location: syntax.Location, *nodes: ast.expr | ast.stmt) -> None:
    for node in nodes:
        node.lineno = location.start.line
        node.col_offset = location.start.column
    if location.end:
        for node in nodes:
            node.end_lineno = location.end.line
            node.end_col_offset = location.end.column


class AstGenerator:
    def generate_ast(self, term: syntax.Term, setting: syntax.BackendSetting) -> ast.AST:
        if (ast_node := self.try_generate_ast(term, setting)) is not None:
            return ast_node
        if (unfolded := term.unfold()) and unfolded is not term:
            return self.generate_ast(unfolded, setting)
        raise tapl_error.TaplError(
            f'The python_backend does not support AST generation for term: {term.__class__.__name__}'
        )

    def generate_stmt(self, term: syntax.Term, setting: syntax.BackendSetting) -> list[ast.stmt]:
        if (stmts := self.try_generate_stmt(term, setting)) is not None:
            return stmts
        if (unfolded := term.unfold()) and unfolded is not term:
            return self.generate_stmt(unfolded, setting)
        raise tapl_error.TaplError(
            f'The python_backend does not support statement generation for term: {term.__class__.__name__}'
        )

    def generate_expr(self, term: syntax.Term, setting: syntax.BackendSetting) -> ast.expr:
        if (expr := self.try_generate_expr(term, setting)) is not None:
            return expr
        if (unfolded := term.unfold()) and unfolded is not term:
            return self.generate_expr(unfolded, setting)
        raise tapl_error.TaplError(
            f'The python_backend does not support expression generation for term: {term.__class__.__name__}'
        )

    def try_generate_ast(self, term: syntax.Term, setting: syntax.BackendSetting) -> ast.AST | None:
        if isinstance(term, terms.Module):
            stmts = []
            for t in term.body:
                stmts.extend(self.generate_stmt(t, setting))
            return ast.Module(body=stmts, type_ignores=[])

        if isinstance(term, syntax.BackendSettingTerm):
            new_setting = term.new_setting(setting)
            return self.generate_ast(term.term, new_setting)

        return None

    def try_generate_stmt(self, term: syntax.Term, setting: syntax.BackendSetting) -> list[ast.stmt] | None:
        if isinstance(term, syntax.TermList):
            if term.is_placeholder:
                raise tapl_error.TaplError('The placeholder list must be initialized before code generation.')
            stmts: list[ast.stmt] = []
            for t in term.flattened():
                stmts.extend(self.generate_stmt(t, setting))
            return stmts

        if term is syntax.Empty:
            return []

        if isinstance(term, terms.FunctionDef):
            name = term.name(setting) if callable(term.name) else term.name
            func_def = ast.FunctionDef(
                name=name,
                args=ast.arguments(
                    posonlyargs=[ast.arg(arg=name) for name in term.posonlyargs],
                    args=[ast.arg(arg=name) for name in term.args],
                    vararg=ast.arg(arg=term.vararg) if term.vararg else None,
                    kwonlyargs=[ast.arg(arg=name) for name in term.kwonlyargs],
                    kw_defaults=[self.generate_expr(t, setting) for t in term.kw_defaults],
                    kwarg=ast.arg(arg=term.kwarg) if term.kwarg else None,
                    defaults=[self.generate_expr(t, setting) for t in term.defaults],
                ),
                body=self.generate_stmt(term.body, setting),
                decorator_list=[],
                returns=None,
                type_comment=None,
            )
            locate(term.location, func_def)
            return [func_def]

        if isinstance(term, terms.ClassDef):
            name = term.name(setting) if callable(term.name) else term.name
            class_def = ast.ClassDef(
                name=name,
                bases=[self.generate_expr(b, setting) for b in term.bases],
                keywords=[ast.keyword(arg=k, value=self.generate_expr(v, setting)) for k, v in term.keywords],
                body=[],
                decorator_list=[self.generate_expr(d, setting) for d in term.decorator_list],
            )
            locate(term.location, class_def)
            class_def.body = self.generate_stmt(term.body, setting)
            return [class_def]

        if isinstance(term, terms.Return):
            return_stmt = ast.Return(self.generate_expr(term.value, setting)) if term.value else ast.Return()
            locate(term.location, return_stmt)
            return [return_stmt]

        if isinstance(term, terms.Delete):
            delete_stmt = ast.Delete(
                targets=[self.generate_expr(t, setting) for t in term.targets],
            )
            locate(term.location, delete_stmt)
            return [delete_stmt]

        if isinstance(term, terms.Assign):
            assign_stmt = ast.Assign(
                targets=[self.generate_expr(t, setting) for t in term.targets],
                value=self.generate_expr(term.value, setting),
            )
            locate(term.location, assign_stmt)
            return [assign_stmt]

        if isinstance(term, terms.For):
            for_stmt = ast.For(
                target=self.generate_expr(term.target, setting),
                iter=self.generate_expr(term.iter, setting),
                body=self.generate_stmt(term.body, setting),
                orelse=self.generate_stmt(term.orelse, setting) if term.orelse else [],
                type_comment=None,
            )
            locate(term.location, for_stmt)
            return [for_stmt]

        if isinstance(term, terms.While):
            while_stmt = ast.While(
                test=self.generate_expr(term.test, setting),
                body=self.generate_stmt(term.body, setting),
                orelse=self.generate_stmt(term.orelse, setting),
            )
            locate(term.location, while_stmt)
            return [while_stmt]

        if isinstance(term, terms.If):
            if_stmt = ast.If(
                test=self.generate_expr(term.test, setting),
                body=self.generate_stmt(term.body, setting),
                orelse=self.generate_stmt(term.orelse, setting),
            )
            locate(term.location, if_stmt)
            return [if_stmt]

        if isinstance(term, terms.With):
            items = []
            for item in term.items:
                if isinstance(item, terms.WithItem):
                    items.append(
                        ast.withitem(
                            context_expr=self.generate_expr(item.context_expr, setting),
                            optional_vars=None
                            if item.optional_vars is syntax.Empty
                            else self.generate_expr(item.optional_vars, setting),
                        )
                    )
                else:
                    raise tapl_error.TaplError(
                        f'Unsupported with item type: {item.__class__.__name__} in python backend.'
                    )
            with_stmt = ast.With(
                items=items,
                body=self.generate_stmt(term.body, setting),
                type_comment=None,
            )
            locate(term.location, with_stmt)
            return [with_stmt]

        if isinstance(term, terms.Try):
            handlers = []
            for h in term.handlers:
                if isinstance(h, terms.ExceptHandler):
                    optional_name = h.name(setting) if callable(h.name) else h.name
                    handlers.append(
                        ast.ExceptHandler(
                            type=self.generate_expr(h.exception_type, setting) if h.exception_type else None,
                            name=optional_name,
                            body=self.generate_stmt(h.body, setting),
                        )
                    )
                else:
                    raise tapl_error.TaplError(
                        f'Unsupported except handler type: {h.__class__.__name__} in python backend.'
                    )
            try_stmt = ast.Try(
                body=self.generate_stmt(term.body, setting),
                handlers=handlers,
                orelse=self.generate_stmt(term.orelse, setting),
                finalbody=self.generate_stmt(term.finalbody, setting),
            )
            locate(term.location, try_stmt)
            return [try_stmt]

        if isinstance(term, terms.Import):
            import_stmt = ast.Import(names=[ast.alias(name=n.name, asname=n.asname) for n in term.names])
            locate(term.location, import_stmt)
            return [import_stmt]

        if isinstance(term, terms.ImportFrom):
            import_from = ast.ImportFrom(
                module=term.module,
                names=[ast.alias(name=n.name, asname=n.asname) for n in term.names],
                level=term.level,
            )
            locate(term.location, import_from)
            return [import_from]

        if isinstance(term, terms.Expr):
            expr_stmt = ast.Expr(value=self.generate_expr(term.value, setting))
            locate(term.location, expr_stmt)
            return [expr_stmt]

        if isinstance(term, syntax.BackendSettingTerm):
            new_setting = term.new_setting(setting)
            return self.generate_stmt(term.term, new_setting)

        if isinstance(term, terms.Pass):
            pass_stmt = ast.Pass()
            locate(term.location, pass_stmt)
            return [pass_stmt]

        return None

    def try_generate_expr(self, term: syntax.Term, setting: syntax.BackendSetting) -> ast.expr | None:
        if isinstance(term, terms.BoolOp):
            bool_op = ast.BoolOp(
                op=BOOL_OP_MAP[term.operator], values=[self.generate_expr(v, setting) for v in term.values]
            )
            locate(term.location, bool_op)
            return bool_op

        if isinstance(term, terms.BinOp):
            bin_op = ast.BinOp(
                left=self.generate_expr(term.left, setting),
                op=BIN_OP_MAP[term.op],
                right=self.generate_expr(term.right, setting),
            )
            locate(term.location, bin_op)
            return bin_op

        if isinstance(term, terms.UnaryOp):
            op = ast.UnaryOp(op=UNARY_OP_MAP[term.op], operand=self.generate_expr(term.operand, setting))
            locate(term.location, op)
            return op

        if isinstance(term, terms.Dict):
            dict_expr = ast.Dict(
                keys=[self.generate_expr(k, setting) for k in term.keys],
                values=[self.generate_expr(v, setting) for v in term.values],
            )
            locate(term.location, dict_expr)
            return dict_expr

        if isinstance(term, terms.Set):
            set_expr = ast.Set(elts=[self.generate_expr(elt, setting) for elt in term.elements])
            locate(term.location, set_expr)
            return set_expr

        if isinstance(term, terms.Compare):
            compare = ast.Compare(
                left=self.generate_expr(term.left, setting),
                ops=[COMPARE_OP_MAP[op] for op in term.operators],
                comparators=[self.generate_expr(v, setting) for v in term.comparators],
            )
            locate(term.location, compare)
            return compare

        if isinstance(term, terms.Call):
            call = ast.Call(
                func=self.generate_expr(term.func, setting),
                args=[self.generate_expr(arg, setting) for arg in term.args],
                keywords=[ast.keyword(arg=k, value=self.generate_expr(v, setting)) for k, v in term.keywords],
            )
            locate(term.location, call)
            return call

        if isinstance(term, terms.Constant):
            const = ast.Constant(value=term.value)
            locate(term.location, const)
            return const

        if isinstance(term, terms.Attribute):
            attr_name = term.attr(setting) if callable(term.attr) else term.attr
            attr = ast.Attribute(
                value=self.generate_expr(term.value, setting),
                attr=attr_name,
                ctx=EXPR_CONTEXT_MAP[term.ctx],
            )
            locate(term.location, attr)
            return attr

        if isinstance(term, terms.Subscript):
            subscript = ast.Subscript(
                value=self.generate_expr(term.value, setting),
                slice=self.generate_expr(term.slice, setting),
                ctx=EXPR_CONTEXT_MAP[term.ctx],
            )
            locate(term.location, subscript)
            return subscript

        if isinstance(term, terms.Name):
            name_id = term.id(setting) if callable(term.id) else term.id
            name = ast.Name(id=name_id, ctx=EXPR_CONTEXT_MAP[term.ctx])
            locate(term.location, name)
            return name

        if isinstance(term, terms.List):
            list_expr = ast.List(
                elts=[self.generate_expr(elt, setting) for elt in term.elements],
                ctx=EXPR_CONTEXT_MAP[term.ctx],
            )
            locate(term.location, list_expr)
            return list_expr

        if isinstance(term, terms.Tuple):
            tuple_expr = ast.Tuple(
                elts=[self.generate_expr(elt, setting) for elt in term.elements],
                ctx=EXPR_CONTEXT_MAP[term.ctx],
            )
            locate(term.location, tuple_expr)
            return tuple_expr

        if isinstance(term, syntax.BackendSettingTerm):
            new_setting = term.new_setting(setting)
            return self.generate_expr(term.term, new_setting)

        return None



def factorial(n):
    if n == 0 or n == 1:
        return 1
    else:
        return n * factorial(n - 1)

class Dog:

    def __init__(self, name):
        self.name = name

    def bark(self):
        return self.name + ' says Woof! Woof!'

def greet_dog(dog):
    return 'Hello, ' + dog.name + '!'

def make_dog(factory, name):
    return factory(name)

def main():
    print(factorial(5))
    my_dog = Dog('Buddy')
    print(my_dog.bark())
    print(greet_dog(my_dog))
    new_dog = make_dog(Dog, 'Max')
    print(new_dog.bark())
if __name__ == '__main__':
    main()


# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from tapl_lang.lib import kinds

Types = {
    'Any': kinds.Any(),
    'Nothing': kinds.Nothing(),
    'NoneType': kinds.NoneType(),
    'Bool': kinds.Record(fields={}, label='Bool'),
    'Int': kinds.Record(fields={}, label='Int'),
    'Float': kinds.Record(fields={}, label='Float'),
    'Str': kinds.Record(fields={}, label='Str'),
}

Any = Types['Any']
Nothing = Types['Nothing']
NoneType = Types['NoneType']
Bool = Types['Bool']
Int = Types['Int']
Float = Types['Float']
Str = Types['Str']


def _split_args(params):
    posonlyargs = []
    args = []
    tuple_found = False
    for p in params:
        if isinstance(p, tuple):
            tuple_found = True
            args.append((p[0], p[1]))
        elif not tuple_found:
            posonlyargs.append(p)
        else:
            raise ValueError('Positional-only arguments must come before regular arguments')
    return posonlyargs, args


def _init_methods(methods):
    """Create a dict of methods for the type with the given method template.
    methods: dict[str, tuple[list[str|list[str]], str]]: A list of methods for the type, where each method is represented as a tuple
    of (name, parameters, result). parameters are list of types which are represented as string or union of strings.
    """
    fields = {}
    for name, (params, result) in methods.items():
        posonlyargs, args = _split_args(params)
        fields[name] = kinds.Function(posonlyargs=posonlyargs, args=args, result=result)
    return fields


def _init_record(record, methods):
    object.__setattr__(record, '_fields__sa', _init_methods(methods))


_init_record(
    Bool, {'__hash__sa': ([], Int), '__eq__sa': ([Bool], Bool), '__lt__': ([Bool], Bool), '__gt__': ([Bool], Bool)}
)
_init_record(
    Int,
    {
        '__hash__sa': ([], Int),
        '__add__': ([Int], Int),
        '__sub__': ([Int], Int),
        '__mul__': ([Int], Int),
        '__truediv__': ([Int], Float),
        '__mod__': ([Int], Int),
        '__floordiv__': ([Int], Int),
        '__eq__sa': ([Int], Bool),
        '__ne__': ([Int], Bool),
        '__lt__': ([Int], Bool),
        '__gt__': ([Int], Bool),
        '__ge__': ([Int], Bool),
        '__le__': ([Int], Bool),
        '__neg__': ([], Int),
    },
)
_init_record(
    Float,
    {
        '__hash__sa': ([], Int),
        '__add__': ([Float], Float),
        '__sub__': ([Float], Float),
        '__mul__': ([Float], Float),
        '__eq__sa': ([Float], Bool),
        '__ne__': ([Float], Bool),
        '__lt__': ([Float], Bool),
        '__gt__': ([Float], Bool),
        '__ge__': ([Float], Bool),
        '__le__': ([Float], Bool),
        '__neg__': ([], Float),
    },
)
_init_record(
    Str,
    {
        '__hash__sa': ([], Int),
        '__eq__sa': ([Str], Bool),
        'isalpha': ([], Bool),
        'isdigit': ([], Bool),
        '__add__': ([Str], Str),
    },
)


def List(element_type):
    methods = {
        'append': ([element_type], NoneType),
        '__len__': ([], Int),
        '__getitem__': ([Int], element_type),
        '__setitem__': ([Int, element_type], NoneType),
        '__delitem__': ([Int], NoneType),
    }
    return kinds.Record(
        fields=_init_methods(methods),
        label=f'List({element_type})',
    )


def Set(element_type):
    methods = {
        'add': ([element_type], NoneType),
        'remove': ([element_type], NoneType),
        '__len__': ([], Int),
        '__contains__': ([element_type], Bool),
    }
    return kinds.Record(
        fields=_init_methods(methods),
        label=f'Set({element_type})',
    )


def Dict(key_type, value_type):
    methods = {
        'get': ([key_type], value_type),
        'set': ([key_type, value_type], NoneType),
        'delete': ([key_type], NoneType),
        '__len__': ([], Int),
        '__getitem__': ([key_type], value_type),
        '__setitem__': ([key_type, value_type], NoneType),
        '__delitem__': ([key_type], NoneType),
    }
    return kinds.Record(
        fields=_init_methods(methods),
        label=f'Dict({key_type}, {value_type})',
    )


TYPE_CONSTRUCTORS = {
    'List': List,
    'Set': Set,
    'Dict': Dict,
}



from tapl_lang.lib import tapl_dev
from tapl_lang.lib import builtin_types
Int = builtin_types.Int

def simple_id(a):
    return a
tapl_dev.log(simple_id(3))

def poly_id(A):

    def typed_id(a):
        return a
    return typed_id
id_int = poly_id(None)
id_str = poly_id(None)
tapl_dev.log(id_int(3))
tapl_dev.log(tapl_dev.describe(id_str))
tapl_dev.log(id_str('abc'))

def auto_id(a):

    def typed_id(b):
        return b
    return typed_id(a)
tapl_dev.log(auto_id('hello'))

def Slot(T):

    class Slot_:
        class_name = 'Slot({})'.format(T)

        def __init__(self, value):
            self._value = value

        def set(self, value):
            self._value = value

        def get(self):
            return self._value

        def __repr__(self):
            return 'Slot(' + str(self._value) + ')'
    return Slot_
slot = Slot(Int)(5)
tapl_dev.log(slot)
tapl_dev.log(slot.set)



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import ast

from tapl_lang.lib import compiler


def test_simple_pipe():
    [ast1, ast2] = compiler.compile_tapl("""language pipeweaver
-7.3 |> print
""")
    assert ast.unparse(ast1) == """print(-7.3)"""
    assert (
        ast.unparse(ast2)
        == """from tapl_lang.lib import tapl_typing
from tapl_lang.pythonlike.predef1 import predef_scope as predef_scope__sa
s0 = tapl_typing.create_scope(parent__sa=predef_scope__sa)
s0.print(-s0.Float)"""
    )


def test_complex_pipe():
    [ast1, ast2] = compiler.compile_tapl("""language pipeweaver
-7.3 |> round |> abs |> print
""")
    assert ast.unparse(ast1) == """print(abs(round(-7.3)))"""
    assert (
        ast.unparse(ast2)
        == """from tapl_lang.lib import tapl_typing
from tapl_lang.pythonlike.predef1 import predef_scope as predef_scope__sa
s0 = tapl_typing.create_scope(parent__sa=predef_scope__sa)
s0.print(s0.abs(s0.round(-s0.Float)))"""
    )



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import ast
import logging

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


parsed_code = ast.parse(
    """
def a(b=1, /, c=3, d=4, *e, f, **g):
    pass
""",
    mode='exec',
)
logger.info(ast.dump(parsed_code, include_attributes=False, indent=4))
compiled_code = compile(parsed_code, filename='', mode='exec')
logger.info(ast.unparse(parsed_code))
# result = eval(compiled_code)
# logging.info(result)


class MyMeta(type):
    def __call__(cls, *args, **kwargs):
        print('Custom __call__ in metaclass')
        instance = super().__call__(*args, **kwargs)
        print('Instance created')
        return instance


class MyClass(metaclass=MyMeta):
    def __init__(self):
        print('Inside __init__')



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import types

log = print


def describe(value):
    if isinstance(value, types.FunctionType):
        return f'<function {value.__name__}>'
    return str(value)



from tapl_lang.lib import tapl_typing
from tapl_lang.pythonlike.predef1 import predef_scope as predef_scope__sa
s0 = tapl_typing.create_scope(parent__sa=predef_scope__sa)
tapl_typing.import_module(s0, ['tapl_dev'], 'tapl_lang.lib', 0)
tapl_typing.import_module(s0, ['examples.matrix'])
s0.matrix = s0.examples.matrix
s0.matrix.accept_matrix_2_3(s0.matrix.Matrix(2, 3)())
s0.tapl_dev.log(s0.tapl_dev.describe(s0.matrix.accept_matrix_2_3))
s0.tapl_dev.log(s0.tapl_dev.describe(s0.matrix.add(2, 2)))
s0.matrix_2_2 = s0.matrix.Matrix(2, 2)()
s0.matrix_2_2.values = tapl_typing.create_typed_list(tapl_typing.create_typed_list(s0.Int, s0.Int), tapl_typing.create_typed_list(s0.Int, s0.Int))
s0.matrix_2_3 = s0.matrix.Matrix(2, 3)()
s0.matrix_2_3.values = tapl_typing.create_typed_list(tapl_typing.create_typed_list(s0.Int, s0.Int, s0.Int), tapl_typing.create_typed_list(s0.Int, s0.Int, s0.Int))
s0.tapl_dev.log(s0.matrix.add(2, 2)(s0.matrix_2_2, s0.matrix_2_2))
s0.tapl_dev.log(s0.matrix.multiply(2, 2, 3)(s0.matrix_2_2, s0.matrix_2_3))
s0.tapl_dev.log(s0.Str)



from tapl_lang.lib import tapl_typing
from tapl_lang.pythonlike.predef1 import predef_scope as predef_scope__sa
s0 = tapl_typing.create_scope(parent__sa=predef_scope__sa)

def factorial(n):
    s1 = tapl_typing.create_scope(parent__sa=s0, n=n)
    s1.factorial = tapl_typing.create_function([s1.Int], s1.Int)
    tapl_typing.set_return_type(s1, s1.Int)
    with tapl_typing.scope_forker(s1) as f1:
        s2 = tapl_typing.fork_scope(f1)
        tapl_typing.create_union(s2.n == s2.Int, s2.n == s2.Int)
        tapl_typing.add_return_type(s2, s2.Int)
        s2 = tapl_typing.fork_scope(f1)
        tapl_typing.add_return_type(s2, s2.n * s2.factorial(s2.n - s2.Int))
    return tapl_typing.get_return_type(s1)
s0.factorial = tapl_typing.create_function([s0.Int], factorial(s0.Int))

class Dog:

    def __init__(self, name):
        s1 = tapl_typing.create_scope(parent__sa=s0, self=self, name=name)
        s1.self.name = s1.name
        return tapl_typing.get_return_type(s1)

    def bark(self):
        s1 = tapl_typing.create_scope(parent__sa=s0, self=self)
        s1.bark = tapl_typing.create_function([], s1.Str)
        tapl_typing.set_return_type(s1, s1.Str)
        tapl_typing.add_return_type(s1, s1.self.name + s1.Str)
        return tapl_typing.get_return_type(s1)
s0.Dog = tapl_typing.create_class(cls=Dog, init_args=[s0.Str], methods=[('bark', [])])

def greet_dog(dog):
    s1 = tapl_typing.create_scope(parent__sa=s0, dog=dog)
    s1.greet_dog = tapl_typing.create_function([s1.Dog.result__sa], s1.Str)
    tapl_typing.set_return_type(s1, s1.Str)
    tapl_typing.add_return_type(s1, s1.Str + s1.dog.name + s1.Str)
    return tapl_typing.get_return_type(s1)
s0.greet_dog = tapl_typing.create_function([s0.Dog.result__sa], greet_dog(s0.Dog.result__sa))

def make_dog(factory, name):
    s1 = tapl_typing.create_scope(parent__sa=s0, factory=factory, name=name)
    s1.make_dog = tapl_typing.create_function([s1.Dog, s1.Str], s1.Dog.result__sa)
    tapl_typing.set_return_type(s1, s1.Dog.result__sa)
    tapl_typing.add_return_type(s1, s1.factory(s1.name))
    return tapl_typing.get_return_type(s1)
s0.make_dog = tapl_typing.create_function([s0.Dog, s0.Str], make_dog(s0.Dog, s0.Str))

def main():
    s1 = tapl_typing.create_scope(parent__sa=s0)
    s1.print(s1.factorial(s1.Int))
    s1.my_dog = s1.Dog(s1.Str)
    s1.print(s1.my_dog.bark())
    s1.print(s1.greet_dog(s1.my_dog))
    s1.new_dog = s1.make_dog(s1.Dog, s1.Str)
    s1.print(s1.new_dog.bark())
    return tapl_typing.get_return_type(s1)
s0.main = tapl_typing.create_function([], main())
with tapl_typing.scope_forker(s0) as f0:
    s1 = tapl_typing.fork_scope(f0)
    s1.__name__ == s1.Str
    s1.main()
    s1 = tapl_typing.fork_scope(f0)


from tapl_lang.lib import tapl_dev
x = 42
tapl_dev.log(x)
a = b = 'hello'
tapl_dev.log(a)
tapl_dev.log(b)
f = 0.1
try:
    f = 1 / 0
except ZeroDivisionError:
    tapl_dev.log('Caught division by zero!')
finally:
    tapl_dev.log('Finished try-except block.')
c = [1, 2, 3]
tapl_dev.log(c)
x = c[1]
tapl_dev.log(c[1])
c[2] = 42
del c[0]
tapl_dev.log(c)
s = {1, 2, 2, 3}
tapl_dev.log(s)
if 2 in s:
    tapl_dev.log('2 is in the set')
s.add(4)
s.remove(1)
tapl_dev.log(s)
dict_obj = {'key1': 'value1', 'key2': 'value2'}
tapl_dev.log(dict_obj)
value = dict_obj['key1']
tapl_dev.log(dict_obj['key1'])
dict_obj['key2'] = 'new_value2'
del dict_obj['key1']
tapl_dev.log(dict_obj)
print('Double quotes string')



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from tapl_language.pipeweaver import pipeweaver_language


def get_language():
    return pipeweaver_language.PipeweaverLanguage()



def Matrix(rows, cols):

    class Matrix_:
        class_name = 'Matrix({},{})'.format(rows, cols)

        def __init__(self):
            self.rows = rows
            self.cols = cols
            self.num_rows = rows
            self.num_cols = cols
            self.values = []
            for i in range(self.num_rows):
                columns = []
                for j in range(self.num_cols):
                    columns.append(0)
                self.values.append(columns)

        def __repr__(self):
            return str(self.values)
    return Matrix_

def accept_matrix_2_3(matrix):
    pass

def add(rows, cols):

    def add_(a, b):
        result = Matrix(rows, cols)()
        for i in range(result.num_rows):
            for j in range(result.num_cols):
                result.values[i][j] = a.values[i][j] + b.values[i][j]
        return result
    return add_

def multiply(m, n, p):

    def multiply_(a, b):
        result = Matrix(m, p)()
        for i in range(a.num_rows):
            for j in range(b.num_cols):
                for k in range(a.num_cols):
                    result.values[i][j] = result.values[i][j] + a.values[i][k] * b.values[k][j]
        return result
    return multiply_

def main():
    matrix_2_2 = Matrix(2, 2)()
    matrix_2_2.values = [[1, 2], [3, 4]]
    matrix_2_3 = Matrix(2, 3)()
    matrix_2_3.values = [[1, 2, 3], [4, 5, 6]]
    accept_matrix_2_3(matrix_2_3)
    print(add(2, 2)(matrix_2_2, matrix_2_2))
    print(multiply(2, 2, 3)(matrix_2_2, matrix_2_3))
if __name__ == '__main__':
    main()


from tapl_lang.lib import tapl_dev
import examples.easy
tapl_dev.log(examples.easy.factorial(7))
my_dog = examples.easy.Dog('Simba')
print(my_dog.bark())
print(examples.easy.greet_dog(my_dog))
examples.easy.main()



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import ast
import importlib
import re

from tapl_lang.core import chunker, syntax, tapl_error
from tapl_lang.lib import python_backend, terms


def gather_errors(term: syntax.Term) -> list[syntax.ErrorTerm]:
    error_bucket: list[syntax.ErrorTerm] = []

    def gather_errors_recursive(t: syntax.Term) -> None:
        if isinstance(t, syntax.ErrorTerm):
            error_bucket.append(t)
        for child in t.children():
            gather_errors_recursive(child)

    gather_errors_recursive(term)
    return error_bucket


def extract_language(chunk: chunker.Chunk) -> str:
    if chunk.children:
        raise tapl_error.TaplError('language clause chunk should not have children.')
    for i in range(1, len(chunk.line_records)):
        if not chunk.line_records[i].empty:
            raise tapl_error.TaplError('language clause chunk should be the first line.')
    pattern = r'^language ([a-zA-Z_][a-zA-Z0-9_]*)$'
    line = chunk.line_records[0].text
    match = re.findall(pattern, line)
    if not match:
        raise tapl_error.TaplError(f'Could not parse language clause[{line}]')
    return match[0]


def make_safe_term(term: syntax.Term) -> syntax.Term:
    return syntax.BackendSettingTerm(
        backend_setting_changer=syntax.Layers(
            layers=[
                syntax.BackendSettingChanger(changer=lambda _: syntax.BackendSetting(scope_level=0)),
                syntax.BackendSettingChanger(changer=lambda _: syntax.BackendSetting(scope_level=0)),
            ]
        ),
        term=term,
    )


def compile_tapl(text: str) -> list[ast.AST]:
    chunks = chunker.chunk_text(text)
    language_name = extract_language(chunks[0])
    language = importlib.import_module(f'tapl_language.{language_name}').get_language()
    predef_headers = language.get_predef_headers()
    predef_layers = syntax.Layers(predef_headers)
    module = terms.Module(body=[predef_layers, syntax.TermList(terms=[], is_placeholder=True)])
    language.parse_chunks(chunks[1:], [module])
    error_bucket: list[syntax.ErrorTerm] = gather_errors(module)
    if error_bucket:
        messages = [repr(e) for e in error_bucket]
        raise tapl_error.TaplError(f'{len(error_bucket)} parsing error(s) found:\n\n' + '\n\n'.join(messages))
    safe_module = make_safe_term(module)
    ls = syntax.LayerSeparator(len(predef_layers.layers))
    layers = ls.build(lambda layer: layer(safe_module))
    return [python_backend.AstGenerator().generate_ast(layer, syntax.BackendSetting(scope_level=0)) for layer in layers]



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from tapl_lang.lib import dynamic_attribute, kinds, scope
from tapl_lang.pythonlike import builtins


def create_function(parameters, result):
    func = kinds.Function(parameters=parameters, result=result)
    return dynamic_attribute.DynamicAttributeMixin(func)


predef_scope = scope.Scope()
predef_scope.store_many__sa(builtins.export1)



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from tapl_lang.lib import scope, tapl_dev

s0 = scope.Scope(fields={'log': tapl_dev.log, 'describe': tapl_dev.describe}, label__sa=__name__)



from tapl_lang.lib import tapl_typing
from tapl_lang.pythonlike.predef1 import predef_scope as predef_scope__sa
s0 = tapl_typing.create_scope(parent__sa=predef_scope__sa)
s0.i = s0.Int
s0.i = s0.Int
s0.sum = s0.Int
with tapl_typing.scope_forker(s0) as f0:
    s1 = tapl_typing.fork_scope(f0)
    s1.i > s1.Int
    s1.sum = s1.sum + s1.i
    s1.i = s1.i - s1.Int
    s1 = tapl_typing.fork_scope(f0)
s0.print(s0.sum)
s0.sum = s0.Int
with tapl_typing.scope_forker(s0) as f0:
    s1 = tapl_typing.fork_scope(f0)
    s1.i = s1.range(s1.Int).__iter__().__next__()
    s1.sum = s1.sum + s1.i
    s1 = tapl_typing.fork_scope(f0)
s0.print(s0.sum)



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception


from abc import ABC, abstractmethod

from tapl_lang.core import chunker, parser, syntax, tapl_error


class Language(ABC):
    def parse_chunks(self, chunks: list[chunker.Chunk], parent_stack: list[syntax.Term]) -> None:
        delayed_statements: syntax.TermList | None = syntax.find_placeholder(parent_stack[-1])
        if delayed_statements is None:
            raise tapl_error.TaplError(
                f'The top of parent_stack[{parent_stack[-1].__class__.__name__}] does not have a placeholder to hold parsed child terms.'
            )
        body: list[syntax.Term] = []
        for chunk in chunks:
            term = self.parse_chunk(chunk, parent_stack)
            if isinstance(term, syntax.SiblingTerm):
                term.integrate_into(body)
            else:
                body.append(term)
        delayed_statements.terms = body
        delayed_statements.is_placeholder = False

    def parse_chunk(self, chunk: chunker.Chunk, parent_stack: list[syntax.Term]) -> syntax.Term:
        grammar = self.get_grammar(parent_stack)
        term = parser.parse_line_records(chunk.line_records, grammar)
        if not isinstance(term, syntax.ErrorTerm) and chunk.children:
            parent_stack.append(term)
            try:
                self.parse_chunks(chunk.children, parent_stack)
            finally:
                parent_stack.pop()
        return term

    @abstractmethod
    def get_grammar(self, parent_stack: list[syntax.Term]) -> parser.Grammar:
        """Returns the grammar for the language."""

    @abstractmethod
    def get_predef_headers(self) -> list[syntax.Term]:
        """Returns the list of each layer's predefined headers for the language."""



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception


from tapl_lang.core import parser, syntax
from tapl_lang.lib import terms
from tapl_lang.pythonlike import grammar
from tapl_lang.pythonlike import rule_names as rn


def parse_expr(text: str, start_rule: str, *, mode: syntax.Term = terms.MODE_SAFE, debug=False) -> syntax.Term:
    return parser.parse_text(
        text,
        grammar=parser.Grammar(grammar.get_grammar().rule_map, start_rule),
        debug=debug,
        config=parser.Config(mode=mode),
    )


def create_loc(start_line: int, start_col: int, end_line: int, end_col: int) -> syntax.Location:
    return syntax.Location(
        start=syntax.Position(line=start_line, column=start_col), end=syntax.Position(line=end_line, column=end_col)
    )


def test_compound_stmt__evaluate_only():
    actual = parse_expr('EVALUATE_ONLY:', rn.COMPOUND_STMT)
    expected = terms.LayerOnly(layer_index=0, term=syntax.TermList(terms=[], is_placeholder=True))
    assert actual == expected


def test_compound_stmt__typecheck_only():
    actual = parse_expr('TYPECHECK_ONLY:', rn.COMPOUND_STMT)
    expected = terms.LayerOnly(layer_index=1, term=syntax.TermList(terms=[], is_placeholder=True))
    assert actual == expected


def test_compound_stmt__function_def():
    actual = parse_expr('def add(x: Int, y: Int) -> Int:', rn.COMPOUND_STMT, mode=terms.MODE_EVALUATE)
    expected = terms.TypedFunctionDef(
        location=create_loc(1, 0, 1, 31),
        name='add',
        parameters=[
            terms.Parameter(
                location=create_loc(1, 8, 1, 14),
                name='x',
                type_=syntax.Layers(
                    layers=[
                        syntax.Empty,
                        terms.TypedName(
                            location=create_loc(1, 11, 1, 14), id='Int', ctx='load', mode=terms.MODE_TYPECHECK
                        ),
                    ]
                ),
                default=syntax.Empty,
                mode=terms.MODE_EVALUATE,
                category=terms.ParamCategory.REGULAR,
            ),
            terms.Parameter(
                location=create_loc(1, 15, 1, 22),
                name='y',
                type_=syntax.Layers(
                    layers=[
                        syntax.Empty,
                        terms.TypedName(
                            location=create_loc(1, 19, 1, 22), id='Int', ctx='load', mode=terms.MODE_TYPECHECK
                        ),
                    ]
                ),
                default=syntax.Empty,
                mode=terms.MODE_EVALUATE,
                category=terms.ParamCategory.REGULAR,
            ),
        ],
        return_type=terms.TypedName(location=create_loc(1, 27, 1, 30), id='Int', ctx='load', mode=terms.MODE_TYPECHECK),
        body=syntax.TermList(terms=[], is_placeholder=True),
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_assignment__annotated():
    actual = parse_expr('x: Int = 42', rn.ASSIGNMENT, mode=terms.MODE_EVALUATE)
    expected = terms.TypedAssign(
        location=create_loc(1, 0, 1, 11),
        target_name=terms.TypedName(location=create_loc(1, 0, 1, 1), id='x', ctx='store', mode=terms.MODE_EVALUATE),
        target_type=terms.TypedName(location=create_loc(1, 3, 1, 6), id='Int', ctx='load', mode=terms.MODE_TYPECHECK),
        value=terms.IntegerLiteral(value=42, mode=terms.MODE_EVALUATE, location=create_loc(1, 9, 1, 11)),
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_assignment__multi_targets():
    actual = parse_expr('a = b = c = 1', rn.ASSIGNMENT, mode=terms.MODE_EVALUATE)
    expected = terms.Assign(
        location=create_loc(1, 0, 1, 13),
        targets=[
            terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='store', mode=terms.MODE_EVALUATE),
            terms.TypedName(location=create_loc(1, 4, 1, 5), id='b', ctx='store', mode=terms.MODE_EVALUATE),
            terms.TypedName(location=create_loc(1, 8, 1, 9), id='c', ctx='store', mode=terms.MODE_EVALUATE),
        ],
        value=terms.IntegerLiteral(value=1, mode=terms.MODE_EVALUATE, location=create_loc(1, 12, 1, 13)),
    )
    assert actual == expected


def test_assignment__no_annotation():
    actual = parse_expr('x = 42', rn.ASSIGNMENT, mode=terms.MODE_EVALUATE)
    expected = terms.Assign(
        location=create_loc(1, 0, 1, 6),
        targets=[
            terms.TypedName(location=create_loc(1, 0, 1, 1), id='x', ctx='store', mode=terms.MODE_EVALUATE),
        ],
        value=terms.IntegerLiteral(value=42, mode=terms.MODE_EVALUATE, location=create_loc(1, 4, 1, 6)),
    )
    assert actual == expected


def test_simple_stmt__assignment():
    actual = parse_expr('x = 10', rn.SIMPLE_STMT, mode=terms.MODE_EVALUATE)
    expected = terms.Assign(
        location=create_loc(1, 0, 1, 6),
        targets=[
            terms.TypedName(location=create_loc(1, 0, 1, 1), id='x', ctx='store', mode=terms.MODE_EVALUATE),
        ],
        value=terms.IntegerLiteral(value=10, mode=terms.MODE_EVALUATE, location=create_loc(1, 4, 1, 6)),
    )
    assert actual == expected


def test_simple_stmt__star_expressions():
    actual = parse_expr('a, b, c', rn.SIMPLE_STMT, mode=terms.MODE_EVALUATE)
    expected = terms.Expr(
        location=create_loc(1, 0, 1, 7),
        value=terms.Tuple(
            location=create_loc(1, 0, 1, 7),
            elements=[
                terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE),
                terms.TypedName(location=create_loc(1, 3, 1, 4), id='b', ctx='load', mode=terms.MODE_EVALUATE),
                terms.TypedName(location=create_loc(1, 6, 1, 7), id='c', ctx='load', mode=terms.MODE_EVALUATE),
            ],
            ctx='load',
        ),
    )
    assert actual == expected


def test_simple_stmt__return():
    actual = parse_expr('return x + 1', rn.SIMPLE_STMT, mode=terms.MODE_EVALUATE)
    expected = terms.TypedReturn(
        location=create_loc(1, 0, 1, 12),
        value=terms.BinOp(
            location=create_loc(1, 6, 1, 12),
            left=terms.TypedName(location=create_loc(1, 7, 1, 8), id='x', ctx='load', mode=terms.MODE_EVALUATE),
            op='+',
            right=terms.IntegerLiteral(value=1, mode=terms.MODE_EVALUATE, location=create_loc(1, 11, 1, 12)),
        ),
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_simple_stmt__import_name():
    actual = parse_expr('import math, sys as system', rn.SIMPLE_STMT, mode=terms.MODE_EVALUATE)
    expected = terms.TypedImport(
        location=create_loc(1, 0, 1, 26),
        names=[
            terms.Alias(name='math', asname=None),
            terms.Alias(name='sys', asname='system'),
        ],
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_simple_stmt__raise():
    actual = parse_expr("raise ValueError('Invalid value')", rn.SIMPLE_STMT, mode=terms.MODE_EVALUATE)
    expected = terms.Raise(
        location=create_loc(1, 0, 1, 33),
        exception=terms.Call(
            location=create_loc(1, 5, 1, 33),
            func=terms.TypedName(
                location=create_loc(1, 6, 1, 16), id='ValueError', ctx='load', mode=terms.MODE_EVALUATE
            ),
            args=[
                terms.StringLiteral(value='Invalid value', mode=terms.MODE_EVALUATE, location=create_loc(1, 17, 1, 32)),
            ],
            keywords=[],
        ),
        cause=syntax.Empty,
    )
    assert actual == expected


def test_simple_stmt__pass():
    actual = parse_expr('pass', rn.SIMPLE_STMT)
    expected = terms.Pass(location=create_loc(1, 0, 1, 4))
    assert actual == expected


def test_simple_stmt__del():
    actual = parse_expr('del my_var', rn.SIMPLE_STMT, mode=terms.MODE_EVALUATE)
    expected = terms.Delete(
        location=create_loc(1, 0, 1, 10),
        targets=[
            terms.TypedName(location=create_loc(1, 4, 1, 10), id='my_var', ctx='delete', mode=terms.MODE_EVALUATE)
        ],
    )
    assert actual == expected


def test_import_name__single():
    actual = parse_expr('import module_name', rn.IMPORT_NAME, mode=terms.MODE_EVALUATE)
    expected = terms.TypedImport(
        location=create_loc(1, 0, 1, 18),
        names=[terms.Alias(name='module_name', asname=None)],
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_import_name__multiple():
    actual = parse_expr('import mod1, mod2 as m2, mod3', rn.IMPORT_NAME, mode=terms.MODE_EVALUATE)
    expected = terms.TypedImport(
        location=create_loc(1, 0, 1, 29),
        names=[
            terms.Alias(name='mod1', asname=None),
            terms.Alias(name='mod2', asname='m2'),
            terms.Alias(name='mod3', asname=None),
        ],
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_import_from__path():
    actual = parse_expr('import a.b.c', rn.IMPORT_NAME, mode=terms.MODE_EVALUATE)
    expected = terms.TypedImport(
        location=create_loc(1, 0, 1, 12),
        names=[terms.Alias(name='a.b.c', asname=None)],
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_import_from__path_multiple():
    actual = parse_expr('import a.b.c as k, d.e', rn.IMPORT_NAME, mode=terms.MODE_EVALUATE)
    expected = terms.TypedImport(
        location=create_loc(1, 0, 1, 22),
        names=[
            terms.Alias(name='a.b.c', asname='k'),
            terms.Alias(name='d.e', asname=None),
        ],
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_import_from__simple():
    actual = parse_expr('from os import path', rn.IMPORT_FROM)
    expected = terms.TypedImportFrom(
        location=create_loc(1, 0, 1, 19),
        module='os',
        names=[terms.Alias(name='path', asname=None)],
        level=0,
        mode=terms.MODE_SAFE,
    )
    assert actual == expected


def test_import_from__multiple_names():
    actual = parse_expr('from os import path, getcwd as cwd', rn.IMPORT_FROM)
    expected = terms.TypedImportFrom(
        location=create_loc(1, 0, 1, 34),
        module='os',
        names=[
            terms.Alias(name='path', asname=None),
            terms.Alias(name='getcwd', asname='cwd'),
        ],
        level=0,
        mode=terms.MODE_SAFE,
    )
    assert actual == expected


def test_import_from__dotted_module():
    actual = parse_expr('from os.path import join', rn.IMPORT_FROM)
    expected = terms.TypedImportFrom(
        location=create_loc(1, 0, 1, 24),
        module='os.path',
        names=[terms.Alias(name='join', asname=None)],
        level=0,
        mode=terms.MODE_SAFE,
    )
    assert actual == expected


def test_import_from__relative_single_dot():
    actual = parse_expr('from . import utils', rn.IMPORT_FROM)
    expected = terms.TypedImportFrom(
        location=create_loc(1, 0, 1, 19),
        module=None,
        names=[terms.Alias(name='utils', asname=None)],
        level=1,
        mode=terms.MODE_SAFE,
    )
    assert actual == expected


def test_import_from__relative_double_dot():
    actual = parse_expr('from .. import helpers', rn.IMPORT_FROM)
    expected = terms.TypedImportFrom(
        location=create_loc(1, 0, 1, 22),
        module=None,
        names=[terms.Alias(name='helpers', asname=None)],
        level=2,
        mode=terms.MODE_SAFE,
    )
    assert actual == expected


def test_import_from__relative_ellipsis():
    actual = parse_expr('from ... import base', rn.IMPORT_FROM)
    expected = terms.TypedImportFrom(
        location=create_loc(1, 0, 1, 20),
        module=None,
        names=[terms.Alias(name='base', asname=None)],
        level=3,
        mode=terms.MODE_SAFE,
    )
    assert actual == expected


def test_import_from__relative_with_module():
    actual = parse_expr('from .models import User', rn.IMPORT_FROM)
    expected = terms.TypedImportFrom(
        location=create_loc(1, 0, 1, 24),
        module='models',
        names=[terms.Alias(name='User', asname=None)],
        level=1,
        mode=terms.MODE_SAFE,
    )
    assert actual == expected


def test_import_from__parenthesized():
    actual = parse_expr('from os import (path, getcwd)', rn.IMPORT_FROM)
    expected = terms.TypedImportFrom(
        location=create_loc(1, 0, 1, 29),
        module='os',
        names=[
            terms.Alias(name='path', asname=None),
            terms.Alias(name='getcwd', asname=None),
        ],
        level=0,
        mode=terms.MODE_SAFE,
    )
    assert actual == expected


def test_import_from__parenthesized_trailing_comma():
    actual = parse_expr('from os import (path, getcwd,)', rn.IMPORT_FROM)
    expected = terms.TypedImportFrom(
        location=create_loc(1, 0, 1, 30),
        module='os',
        names=[
            terms.Alias(name='path', asname=None),
            terms.Alias(name='getcwd', asname=None),
        ],
        level=0,
        mode=terms.MODE_SAFE,
    )
    assert actual == expected


def test_import_from__via_import_stmt():
    actual = parse_expr('from sys import argv', rn.IMPORT_STMT)
    expected = terms.TypedImportFrom(
        location=create_loc(1, 0, 1, 20),
        module='sys',
        names=[terms.Alias(name='argv', asname=None)],
        level=0,
        mode=terms.MODE_SAFE,
    )
    assert actual == expected


def test_raise__expression():
    actual = parse_expr('raise 42', rn.RAISE_STMT, mode=terms.MODE_EVALUATE)
    expected = terms.Raise(
        location=create_loc(1, 0, 1, 8),
        exception=terms.IntegerLiteral(value=42, mode=terms.MODE_EVALUATE, location=create_loc(1, 6, 1, 8)),
        cause=syntax.Empty,
    )
    assert actual == expected


def test_raise__no_expression():
    actual = parse_expr('raise', rn.RAISE_STMT)
    expected = terms.Raise(
        location=create_loc(1, 0, 1, 5),
        exception=syntax.Empty,
        cause=syntax.Empty,
    )
    assert actual == expected


def test_del_stmt__single_target():
    actual = parse_expr('del x', rn.DEL_STMT, mode=terms.MODE_EVALUATE)
    expected = terms.Delete(
        location=create_loc(1, 0, 1, 5),
        targets=[terms.TypedName(location=create_loc(1, 4, 1, 5), id='x', ctx='delete', mode=terms.MODE_EVALUATE)],
    )
    assert actual == expected


def test_del_stmt__multi_targets():
    actual = parse_expr('del a, b, c', rn.DEL_STMT, mode=terms.MODE_EVALUATE)
    expected = terms.Delete(
        location=create_loc(1, 0, 1, 11),
        targets=[
            terms.TypedName(location=create_loc(1, 4, 1, 5), id='a', ctx='delete', mode=terms.MODE_EVALUATE),
            terms.TypedName(location=create_loc(1, 7, 1, 8), id='b', ctx='delete', mode=terms.MODE_EVALUATE),
            terms.TypedName(location=create_loc(1, 10, 1, 11), id='c', ctx='delete', mode=terms.MODE_EVALUATE),
        ],
    )
    assert actual == expected


def test_del_stmt__attribute():
    actual = parse_expr('del obj.attr', rn.DEL_STMT, mode=terms.MODE_EVALUATE)
    expected = terms.Delete(
        location=create_loc(1, 0, 1, 12),
        targets=[
            terms.Attribute(
                location=create_loc(1, 3, 1, 12),
                value=terms.TypedName(location=create_loc(1, 4, 1, 7), id='obj', ctx='load', mode=terms.MODE_EVALUATE),
                attr='attr',
                ctx='delete',
            )
        ],
    )
    assert actual == expected


def test_del_stmt__subscript():
    actual = parse_expr('del d[1]', rn.DEL_STMT, mode=terms.MODE_EVALUATE)
    expected = terms.Delete(
        location=create_loc(1, 0, 1, 8),
        targets=[
            terms.Subscript(
                location=create_loc(1, 3, 1, 8),
                value=terms.TypedName(location=create_loc(1, 4, 1, 5), id='d', ctx='load', mode=terms.MODE_EVALUATE),
                slice=terms.IntegerLiteral(value=1, mode=terms.MODE_EVALUATE, location=create_loc(1, 6, 1, 7)),
                ctx='delete',
            )
        ],
    )
    assert actual == expected


def test_star_targets__single():
    actual = parse_expr('variable', rn.STAR_TARGETS, mode=terms.MODE_EVALUATE)
    expected = terms.TypedName(location=create_loc(1, 0, 1, 8), id='variable', ctx='store', mode=terms.MODE_EVALUATE)
    assert actual == expected


def test_star_targets__multi():
    actual = parse_expr('var1, var2, var3', rn.STAR_TARGETS, mode=terms.MODE_EVALUATE)
    expected = terms.Tuple(
        location=create_loc(1, 0, 1, 16),
        elements=[
            terms.TypedName(location=create_loc(1, 0, 1, 4), id='var1', ctx='store', mode=terms.MODE_EVALUATE),
            terms.TypedName(location=create_loc(1, 6, 1, 10), id='var2', ctx='store', mode=terms.MODE_EVALUATE),
            terms.TypedName(location=create_loc(1, 12, 1, 16), id='var3', ctx='store', mode=terms.MODE_EVALUATE),
        ],
        ctx='store',
    )
    assert actual == expected


def tetst_star_target__targt_with_star_atom():
    actual = parse_expr('variable', rn.STAR_TARGET, mode=terms.MODE_EVALUATE)
    expected = terms.TypedName(location=create_loc(1, 0, 1, 8), id='variable', ctx='store', mode=terms.MODE_EVALUATE)
    assert actual == expected


def test_t_lookahead():
    punct = ['(', '[', '.']
    for p in punct:
        actual = parse_expr(p, rn.T_LOOKAHEAD)
        expected = grammar.TokenPunct(location=create_loc(1, 0, 1, 1), value=p)
        assert actual == expected


def test_t_primary__atom_failed():
    actual = parse_expr('variable', rn.T_PRIMARY, mode=terms.MODE_EVALUATE)
    expected = parser.ParseFailed
    assert actual == expected


def test_target_with_star_atom__name():
    actual = parse_expr('my_var', rn.TARGET_WITH_STAR_ATOM, mode=terms.MODE_EVALUATE)
    expected = terms.TypedName(location=create_loc(1, 0, 1, 6), id='my_var', ctx='store', mode=terms.MODE_EVALUATE)
    assert actual == expected


def test_target_with_star_atom__attribute():
    actual = parse_expr('obj.attr', rn.TARGET_WITH_STAR_ATOM, mode=terms.MODE_EVALUATE)
    expected = terms.Attribute(
        location=create_loc(1, 0, 1, 8),
        value=terms.TypedName(location=create_loc(1, 0, 1, 3), id='obj', ctx='load', mode=terms.MODE_EVALUATE),
        attr='attr',
        ctx='store',
    )
    assert actual == expected


def test_target_with_star_atom__attribute_nested():
    actual = parse_expr('obj.sub.attr', rn.TARGET_WITH_STAR_ATOM, mode=terms.MODE_EVALUATE)
    expected = terms.Attribute(
        location=create_loc(1, 0, 1, 12),
        value=terms.Attribute(
            location=create_loc(1, 0, 1, 7),
            value=terms.TypedName(location=create_loc(1, 0, 1, 3), id='obj', ctx='load', mode=terms.MODE_EVALUATE),
            attr='sub',
            ctx='load',
        ),
        attr='attr',
        ctx='store',
    )
    assert actual == expected


def test_target_with_star_atom__subscript():
    actual = parse_expr('d[:-1]', rn.TARGET_WITH_STAR_ATOM, mode=terms.MODE_EVALUATE, debug=True)
    expected = terms.Subscript(
        location=create_loc(1, 0, 1, 6),
        value=terms.TypedName(location=create_loc(1, 0, 1, 1), id='d', ctx='load', mode=terms.MODE_EVALUATE),
        slice=terms.Slice(
            location=create_loc(1, 2, 1, 5),
            lower=syntax.Empty,
            upper=terms.UnaryOp(
                location=create_loc(1, 3, 1, 5),
                op='-',
                operand=terms.IntegerLiteral(value=1, mode=terms.MODE_EVALUATE, location=create_loc(1, 4, 1, 5)),
            ),
            step=syntax.Empty,
        ),
        ctx='load',
    )
    assert actual == expected


def test_target_with_star_atom__subscript_nested():
    actual = parse_expr('matrix[0][1:5]', rn.TARGET_WITH_STAR_ATOM, mode=terms.MODE_EVALUATE)
    expected = terms.Subscript(
        location=create_loc(1, 0, 1, 14),
        value=terms.Subscript(
            location=create_loc(1, 0, 1, 9),
            value=terms.TypedName(location=create_loc(1, 0, 1, 6), id='matrix', ctx='load', mode=terms.MODE_EVALUATE),
            slice=terms.IntegerLiteral(value=0, mode=terms.MODE_EVALUATE, location=create_loc(1, 7, 1, 8)),
            ctx='load',
        ),
        slice=terms.Slice(
            location=create_loc(1, 10, 1, 13),
            lower=terms.IntegerLiteral(value=1, mode=terms.MODE_EVALUATE, location=create_loc(1, 10, 1, 11)),
            upper=terms.IntegerLiteral(value=5, mode=terms.MODE_EVALUATE, location=create_loc(1, 12, 1, 13)),
            step=syntax.Empty,
        ),
        ctx='load',
    )
    assert actual == expected


def test_slices__single():
    actual = parse_expr('a:b', rn.SLICES, mode=terms.MODE_EVALUATE)
    expected = terms.Slice(
        location=create_loc(1, 0, 1, 3),
        lower=terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE),
        upper=terms.TypedName(location=create_loc(1, 2, 1, 3), id='b', ctx='load', mode=terms.MODE_EVALUATE),
        step=syntax.Empty,
    )
    assert actual == expected


def test_slices__single_as_tuple():
    actual = parse_expr('a,', rn.SLICES, mode=terms.MODE_EVALUATE)
    expected = syntax.TermList(
        terms=[terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE)]
    )
    assert actual == expected


def test_slices__multi():
    actual = parse_expr('x:y, 1:10, ::2', rn.SLICES, mode=terms.MODE_EVALUATE)
    expected = syntax.TermList(
        terms=[
            terms.Slice(
                location=create_loc(1, 0, 1, 3),
                lower=terms.TypedName(location=create_loc(1, 0, 1, 1), id='x', ctx='load', mode=terms.MODE_EVALUATE),
                upper=terms.TypedName(location=create_loc(1, 2, 1, 3), id='y', ctx='load', mode=terms.MODE_EVALUATE),
                step=syntax.Empty,
            ),
            terms.Slice(
                location=create_loc(1, 4, 1, 9),
                lower=terms.IntegerLiteral(value=1, mode=terms.MODE_EVALUATE, location=create_loc(1, 5, 1, 6)),
                upper=terms.IntegerLiteral(value=10, mode=terms.MODE_EVALUATE, location=create_loc(1, 7, 1, 9)),
                step=syntax.Empty,
            ),
            terms.Slice(
                location=create_loc(1, 10, 1, 14),
                lower=syntax.Empty,
                upper=syntax.Empty,
                step=terms.IntegerLiteral(value=2, mode=terms.MODE_EVALUATE, location=create_loc(1, 13, 1, 14)),
            ),
        ],
    )
    assert actual == expected


def test_slices__multi_trailing_comma():
    actual = parse_expr('x:y, ::2,', rn.SLICES, mode=terms.MODE_EVALUATE)
    expected = syntax.TermList(
        terms=[
            terms.Slice(
                location=create_loc(1, 0, 1, 3),
                lower=terms.TypedName(location=create_loc(1, 0, 1, 1), id='x', ctx='load', mode=terms.MODE_EVALUATE),
                upper=terms.TypedName(location=create_loc(1, 2, 1, 3), id='y', ctx='load', mode=terms.MODE_EVALUATE),
                step=syntax.Empty,
            ),
            terms.Slice(
                location=create_loc(1, 4, 1, 8),
                lower=syntax.Empty,
                upper=syntax.Empty,
                step=terms.IntegerLiteral(value=2, mode=terms.MODE_EVALUATE, location=create_loc(1, 7, 1, 8)),
            ),
        ],
    )
    assert actual == expected


def test_slice__named_expression():
    actual = parse_expr('x := 5', rn.SLICE, mode=terms.MODE_EVALUATE)
    expected = terms.NamedExpr(
        location=create_loc(1, 0, 1, 6),
        target=terms.TypedName(location=create_loc(1, 0, 1, 1), id='x', ctx='store', mode=terms.MODE_EVALUATE),
        value=terms.IntegerLiteral(value=5, mode=terms.MODE_EVALUATE, location=create_loc(1, 5, 1, 6)),
    )
    assert actual == expected


def test_slice__single_expression():
    actual = parse_expr('y + 2', rn.SLICE, mode=terms.MODE_EVALUATE)
    expected = terms.BinOp(
        location=create_loc(1, 0, 1, 5),
        left=terms.TypedName(location=create_loc(1, 0, 1, 1), id='y', ctx='load', mode=terms.MODE_EVALUATE),
        op='+',
        right=terms.IntegerLiteral(value=2, mode=terms.MODE_EVALUATE, location=create_loc(1, 4, 1, 5)),
    )
    assert actual == expected


def test_slice__range_single():
    actual = parse_expr('1:10', rn.SLICE, mode=terms.MODE_EVALUATE)
    expected = terms.Slice(
        location=create_loc(1, 0, 1, 4),
        lower=terms.IntegerLiteral(value=1, mode=terms.MODE_EVALUATE, location=create_loc(1, 0, 1, 1)),
        upper=terms.IntegerLiteral(value=10, mode=terms.MODE_EVALUATE, location=create_loc(1, 2, 1, 4)),
        step=syntax.Empty,
    )
    assert actual == expected


def test_slice__range_single_no_upper():
    actual = parse_expr('1:', rn.SLICE, mode=terms.MODE_EVALUATE)
    expected = terms.Slice(
        location=create_loc(1, 0, 1, 2),
        lower=terms.IntegerLiteral(value=1, mode=terms.MODE_EVALUATE, location=create_loc(1, 0, 1, 1)),
        upper=syntax.Empty,
        step=syntax.Empty,
    )
    assert actual == expected


def test_slice__range_single_no_lower():
    actual = parse_expr(':10', rn.SLICE, mode=terms.MODE_EVALUATE)
    expected = terms.Slice(
        location=create_loc(1, 0, 1, 3),
        lower=syntax.Empty,
        upper=terms.IntegerLiteral(value=10, mode=terms.MODE_EVALUATE, location=create_loc(1, 1, 1, 3)),
        step=syntax.Empty,
    )
    assert actual == expected


def test_slice__range_full():
    actual = parse_expr('1:10:2', rn.SLICE, mode=terms.MODE_EVALUATE)
    expected = terms.Slice(
        location=create_loc(1, 0, 1, 6),
        lower=terms.IntegerLiteral(value=1, mode=terms.MODE_EVALUATE, location=create_loc(1, 0, 1, 1)),
        upper=terms.IntegerLiteral(value=10, mode=terms.MODE_EVALUATE, location=create_loc(1, 2, 1, 4)),
        step=terms.IntegerLiteral(value=2, mode=terms.MODE_EVALUATE, location=create_loc(1, 5, 1, 6)),
    )
    assert actual == expected


def test_slice__range_no_lower():
    actual = parse_expr(':10:2', rn.SLICE, mode=terms.MODE_EVALUATE)
    expected = terms.Slice(
        location=create_loc(1, 0, 1, 5),
        lower=syntax.Empty,
        upper=terms.IntegerLiteral(value=10, mode=terms.MODE_EVALUATE, location=create_loc(1, 1, 1, 3)),
        step=terms.IntegerLiteral(value=2, mode=terms.MODE_EVALUATE, location=create_loc(1, 4, 1, 5)),
    )
    assert actual == expected


def test_slice__range_no_upper():
    actual = parse_expr('1::2', rn.SLICE, mode=terms.MODE_EVALUATE)
    expected = terms.Slice(
        location=create_loc(1, 0, 1, 4),
        lower=terms.IntegerLiteral(value=1, mode=terms.MODE_EVALUATE, location=create_loc(1, 0, 1, 1)),
        upper=syntax.Empty,
        step=terms.IntegerLiteral(value=2, mode=terms.MODE_EVALUATE, location=create_loc(1, 3, 1, 4)),
    )
    assert actual == expected


def test_slice__range_no_step():
    actual = parse_expr('1:10:', rn.SLICE, mode=terms.MODE_EVALUATE)
    expected = terms.Slice(
        location=create_loc(1, 0, 1, 5),
        lower=terms.IntegerLiteral(value=1, mode=terms.MODE_EVALUATE, location=create_loc(1, 0, 1, 1)),
        upper=terms.IntegerLiteral(value=10, mode=terms.MODE_EVALUATE, location=create_loc(1, 2, 1, 4)),
        step=syntax.Empty,
    )
    assert actual == expected


def test_slice__range_no_lower_upper_step():
    actual = parse_expr('::', rn.SLICE, mode=terms.MODE_EVALUATE)
    expected = terms.Slice(
        location=create_loc(1, 0, 1, 2),
        lower=syntax.Empty,
        upper=syntax.Empty,
        step=syntax.Empty,
    )
    assert actual == expected


def test_slice__range_no_upper_step():
    actual = parse_expr('1::', rn.SLICE, mode=terms.MODE_EVALUATE)
    expected = terms.Slice(
        location=create_loc(1, 0, 1, 3),
        lower=terms.IntegerLiteral(value=1, mode=terms.MODE_EVALUATE, location=create_loc(1, 0, 1, 1)),
        upper=syntax.Empty,
        step=syntax.Empty,
    )
    assert actual == expected


def test_slice__range_no_lower_step():
    actual = parse_expr(':10:', rn.SLICE, mode=terms.MODE_EVALUATE)
    expected = terms.Slice(
        location=create_loc(1, 0, 1, 4),
        lower=syntax.Empty,
        upper=terms.IntegerLiteral(value=10, mode=terms.MODE_EVALUATE, location=create_loc(1, 1, 1, 3)),
        step=syntax.Empty,
    )
    assert actual == expected


def test_slice__range_no_lower_upper():
    actual = parse_expr('::2', rn.SLICE, mode=terms.MODE_EVALUATE)
    expected = terms.Slice(
        location=create_loc(1, 0, 1, 3),
        lower=syntax.Empty,
        upper=syntax.Empty,
        step=terms.IntegerLiteral(value=2, mode=terms.MODE_EVALUATE, location=create_loc(1, 2, 1, 3)),
    )
    assert actual == expected


def test_atom__name():
    actual = parse_expr('variable_name', rn.ATOM, mode=terms.MODE_EVALUATE)
    expected = terms.TypedName(
        location=create_loc(1, 0, 1, 13), id='variable_name', ctx='load', mode=terms.MODE_EVALUATE
    )
    assert actual == expected


def test_atom__true():
    actual = parse_expr('True', rn.ATOM)
    expected = terms.BooleanLiteral(value=True, mode=terms.MODE_SAFE, location=create_loc(1, 0, 1, 4))
    assert actual == expected


def test_atom__lifted_true():
    actual = parse_expr('^True', rn.ATOM)
    expected = terms.BooleanLiteral(value=True, mode=terms.MODE_LIFT, location=create_loc(1, 1, 1, 5))
    assert actual == expected


def test_atom__false():
    actual = parse_expr('False', rn.ATOM)
    expected = terms.BooleanLiteral(value=False, mode=terms.MODE_SAFE, location=create_loc(1, 0, 1, 5))
    assert actual == expected


def test_atom__none():
    actual = parse_expr('None', rn.ATOM)
    expected = terms.NoneLiteral(mode=terms.MODE_SAFE, location=create_loc(1, 0, 1, 4))
    assert actual == expected


def test_atom__string():
    actual = parse_expr("'hello, world!'", rn.ATOM)
    expected = terms.StringLiteral(value='hello, world!', mode=terms.MODE_SAFE, location=create_loc(1, 0, 1, 15))
    assert actual == expected


def test_atom__number_integer():
    actual = parse_expr('42', rn.ATOM)
    expected = terms.IntegerLiteral(value=42, mode=terms.MODE_SAFE, location=create_loc(1, 0, 1, 2))
    assert actual == expected


def test_atom__number_integer_lifted():
    actual = parse_expr('^42', rn.ATOM)
    expected = terms.IntegerLiteral(value=42, mode=terms.MODE_LIFT, location=create_loc(1, 1, 1, 3))
    assert actual == expected


def test_atom__number_float():
    actual = parse_expr('3.14', rn.ATOM)
    expected = terms.FloatLiteral(value=3.14, mode=terms.MODE_SAFE, location=create_loc(1, 0, 1, 4))
    assert actual == expected


def test_group__expression():
    actual = parse_expr('(x + 1)', rn.GROUP, mode=terms.MODE_EVALUATE)
    expected = terms.BinOp(
        location=create_loc(1, 1, 1, 6),
        left=terms.TypedName(location=create_loc(1, 1, 1, 2), id='x', ctx='load', mode=terms.MODE_EVALUATE),
        op='+',
        right=terms.IntegerLiteral(value=1, mode=terms.MODE_EVALUATE, location=create_loc(1, 5, 1, 6)),
    )
    assert actual == expected


def test_group__named_expression():
    actual = parse_expr('(y := 10)', rn.GROUP, mode=terms.MODE_EVALUATE)
    expected = terms.NamedExpr(
        location=create_loc(1, 1, 1, 8),
        target=terms.TypedName(location=create_loc(1, 1, 1, 2), id='y', ctx='store', mode=terms.MODE_EVALUATE),
        value=terms.IntegerLiteral(value=10, mode=terms.MODE_EVALUATE, location=create_loc(1, 6, 1, 8)),
    )
    assert actual == expected


def test_tuple__empty():
    actual = parse_expr('()', rn.ATOM)
    expected = terms.Tuple(
        location=create_loc(1, 0, 1, 2),
        elements=[],
        ctx='load',
    )
    assert actual == expected


def test_tuple__single():
    actual = parse_expr('(42,)', rn.ATOM)
    expected = terms.Tuple(
        location=create_loc(1, 0, 1, 5),
        elements=[
            terms.IntegerLiteral(value=42, mode=terms.MODE_SAFE, location=create_loc(1, 1, 1, 3)),
        ],
        ctx='load',
    )
    assert actual == expected


def test_tuple__multi():
    actual = parse_expr("(1, 'two', None)", rn.ATOM)
    expected = terms.Tuple(
        location=create_loc(1, 0, 1, 16),
        elements=[
            terms.IntegerLiteral(value=1, mode=terms.MODE_SAFE, location=create_loc(1, 1, 1, 2)),
            terms.StringLiteral(value='two', mode=terms.MODE_SAFE, location=create_loc(1, 4, 1, 9)),
            terms.NoneLiteral(mode=terms.MODE_SAFE, location=create_loc(1, 11, 1, 15)),
        ],
        ctx='load',
    )
    assert actual == expected


def test_list__empty():
    # FIXME: convert to MODE_TYPECHECK with unknow type, or No AUTO type inference.
    actual = parse_expr('[]', rn.LIST, mode=terms.MODE_EVALUATE)
    expected = terms.TypedList(
        location=create_loc(1, 0, 1, 2),
        elements=[],
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_list__single():
    actual = parse_expr('[item]', rn.LIST, mode=terms.MODE_EVALUATE)
    expected = terms.TypedList(
        location=create_loc(1, 0, 1, 6),
        elements=[
            terms.TypedName(location=create_loc(1, 1, 1, 5), id='item', ctx='load', mode=terms.MODE_EVALUATE),
        ],
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_list__multi():
    actual = parse_expr('[a, b, c]', rn.LIST, mode=terms.MODE_EVALUATE)
    expected = terms.TypedList(
        location=create_loc(1, 0, 1, 9),
        elements=[
            terms.TypedName(location=create_loc(1, 1, 1, 2), id='a', ctx='load', mode=terms.MODE_EVALUATE),
            terms.TypedName(location=create_loc(1, 4, 1, 5), id='b', ctx='load', mode=terms.MODE_EVALUATE),
            terms.TypedName(location=create_loc(1, 7, 1, 8), id='c', ctx='load', mode=terms.MODE_EVALUATE),
        ],
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_set():
    actual = parse_expr('{x, y, z}', rn.SET, mode=terms.MODE_EVALUATE)
    expected = terms.TypedSet(
        location=create_loc(1, 0, 1, 9),
        elements=[
            terms.TypedName(location=create_loc(1, 1, 1, 2), id='x', ctx='load', mode=terms.MODE_EVALUATE),
            terms.TypedName(location=create_loc(1, 4, 1, 5), id='y', ctx='load', mode=terms.MODE_EVALUATE),
            terms.TypedName(location=create_loc(1, 7, 1, 8), id='z', ctx='load', mode=terms.MODE_EVALUATE),
        ],
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_kvpair():
    actual = parse_expr("'a': 1", rn.KVPAIR, mode=terms.MODE_EVALUATE)
    expected = grammar.KeyValuePair(
        key=terms.StringLiteral(value='a', mode=terms.MODE_EVALUATE, location=create_loc(1, 0, 1, 3)),
        value=terms.IntegerLiteral(value=1, mode=terms.MODE_EVALUATE, location=create_loc(1, 5, 1, 6)),
    )
    assert actual == expected


def test_double_starred_kvpair():
    actual = parse_expr("'a': 1", rn.DOUBLE_STARRED_KVPAIR, mode=terms.MODE_EVALUATE)
    expected = grammar.KeyValuePair(
        key=terms.StringLiteral(value='a', mode=terms.MODE_EVALUATE, location=create_loc(1, 0, 1, 3)),
        value=terms.IntegerLiteral(value=1, mode=terms.MODE_EVALUATE, location=create_loc(1, 5, 1, 6)),
    )
    assert actual == expected


def test_double_starred_kvpair__single():
    actual = parse_expr("'a': 1", rn.DOUBLE_STARRED_KVPAIRS, mode=terms.MODE_EVALUATE)
    expected = syntax.TermList(
        terms=[
            grammar.KeyValuePair(
                key=terms.StringLiteral(value='a', mode=terms.MODE_EVALUATE, location=create_loc(1, 0, 1, 3)),
                value=terms.IntegerLiteral(value=1, mode=terms.MODE_EVALUATE, location=create_loc(1, 5, 1, 6)),
            ),
        ],
    )
    assert actual == expected


def test_double_starred_kvpairs__multi():
    actual = parse_expr("'a': 1, 'b': 2", rn.DOUBLE_STARRED_KVPAIRS, mode=terms.MODE_EVALUATE)
    expected = syntax.TermList(
        terms=[
            grammar.KeyValuePair(
                key=terms.StringLiteral(value='a', mode=terms.MODE_EVALUATE, location=create_loc(1, 0, 1, 3)),
                value=terms.IntegerLiteral(value=1, mode=terms.MODE_EVALUATE, location=create_loc(1, 5, 1, 6)),
            ),
            grammar.KeyValuePair(
                key=terms.StringLiteral(value='b', mode=terms.MODE_EVALUATE, location=create_loc(1, 8, 1, 11)),
                value=terms.IntegerLiteral(value=2, mode=terms.MODE_EVALUATE, location=create_loc(1, 13, 1, 14)),
            ),
        ],
    )
    assert actual == expected


def test_double_starred_kvpairs__trailing_comma():
    actual = parse_expr("'a': 1, 'b': 2,", rn.DOUBLE_STARRED_KVPAIRS, mode=terms.MODE_EVALUATE)
    expected = syntax.TermList(
        terms=[
            grammar.KeyValuePair(
                key=terms.StringLiteral(value='a', mode=terms.MODE_EVALUATE, location=create_loc(1, 0, 1, 3)),
                value=terms.IntegerLiteral(value=1, mode=terms.MODE_EVALUATE, location=create_loc(1, 5, 1, 6)),
            ),
            grammar.KeyValuePair(
                key=terms.StringLiteral(value='b', mode=terms.MODE_EVALUATE, location=create_loc(1, 8, 1, 11)),
                value=terms.IntegerLiteral(value=2, mode=terms.MODE_EVALUATE, location=create_loc(1, 13, 1, 14)),
            ),
        ],
    )
    assert actual == expected


def test_dict__empty():
    actual = parse_expr('{}', rn.DICT, mode=terms.MODE_EVALUATE)
    expected = terms.TypedDict(
        location=create_loc(1, 0, 1, 2),
        keys=[],
        values=[],
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_dict__single():
    actual = parse_expr("{'a': 1}", rn.DICT, mode=terms.MODE_EVALUATE)
    expected = terms.TypedDict(
        location=create_loc(1, 0, 1, 8),
        keys=[terms.StringLiteral(value='a', mode=terms.MODE_EVALUATE, location=create_loc(1, 1, 1, 4))],
        values=[terms.IntegerLiteral(value=1, mode=terms.MODE_EVALUATE, location=create_loc(1, 6, 1, 7))],
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_dict__traling_comma():
    actual = parse_expr("{'a': 1,}", rn.DICT, mode=terms.MODE_EVALUATE)
    expected = terms.TypedDict(
        location=create_loc(1, 0, 1, 9),
        keys=[terms.StringLiteral(value='a', mode=terms.MODE_EVALUATE, location=create_loc(1, 1, 1, 4))],
        values=[terms.IntegerLiteral(value=1, mode=terms.MODE_EVALUATE, location=create_loc(1, 6, 1, 7))],
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_dict__multi():
    actual = parse_expr("{'a': 1, 'b': 2}", rn.DICT, mode=terms.MODE_EVALUATE)
    expected = terms.TypedDict(
        location=create_loc(1, 0, 1, 16),
        keys=[
            terms.StringLiteral(value='a', mode=terms.MODE_EVALUATE, location=create_loc(1, 1, 1, 4)),
            terms.StringLiteral(value='b', mode=terms.MODE_EVALUATE, location=create_loc(1, 9, 1, 12)),
        ],
        values=[
            terms.IntegerLiteral(value=1, mode=terms.MODE_EVALUATE, location=create_loc(1, 6, 1, 7)),
            terms.IntegerLiteral(value=2, mode=terms.MODE_EVALUATE, location=create_loc(1, 14, 1, 15)),
        ],
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_expression__disjunction():
    actual = parse_expr('a or b', rn.EXPRESSION, mode=terms.MODE_SAFE)
    expected = terms.TypedBoolOp(
        location=create_loc(1, 0, 1, 6),
        operator='or',
        values=[
            terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_SAFE),
            terms.TypedName(location=create_loc(1, 5, 1, 6), id='b', ctx='load', mode=terms.MODE_SAFE),
        ],
        mode=terms.MODE_SAFE,
    )
    assert actual == expected


def test_disjunction__single():
    actual = parse_expr('a', rn.DISJUNCTION, mode=terms.MODE_SAFE)
    expected = terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_SAFE)
    assert actual == expected


def test_disjunction__or_chain():
    actual = parse_expr('a or b or c', rn.DISJUNCTION, mode=terms.MODE_EVALUATE)
    expected = terms.TypedBoolOp(
        location=create_loc(1, 0, 1, 11),
        operator='or',
        values=[
            terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE),
            terms.TypedName(location=create_loc(1, 5, 1, 6), id='b', ctx='load', mode=terms.MODE_EVALUATE),
            terms.TypedName(location=create_loc(1, 10, 1, 11), id='c', ctx='load', mode=terms.MODE_EVALUATE),
        ],
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_conjunction__and_chain():
    actual = parse_expr('x and y and z', rn.CONJUNCTION, mode=terms.MODE_EVALUATE)
    expected = terms.TypedBoolOp(
        location=create_loc(1, 0, 1, 13),
        operator='and',
        values=[
            terms.TypedName(location=create_loc(1, 0, 1, 1), id='x', ctx='load', mode=terms.MODE_EVALUATE),
            terms.TypedName(location=create_loc(1, 6, 1, 7), id='y', ctx='load', mode=terms.MODE_EVALUATE),
            terms.TypedName(location=create_loc(1, 12, 1, 13), id='z', ctx='load', mode=terms.MODE_EVALUATE),
        ],
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_conjunction__single():
    actual = parse_expr('x', rn.CONJUNCTION, mode=terms.MODE_EVALUATE)
    expected = terms.TypedName(location=create_loc(1, 0, 1, 1), id='x', ctx='load', mode=terms.MODE_EVALUATE)
    assert actual == expected


def test_inversion__not():
    actual = parse_expr('not flag', rn.INVERSION, mode=terms.MODE_EVALUATE)
    expected = terms.BoolNot(
        location=create_loc(1, 0, 1, 8),
        operand=terms.TypedName(location=create_loc(1, 4, 1, 8), id='flag', ctx='load', mode=terms.MODE_EVALUATE),
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_inversion__single():
    actual = parse_expr('flag', rn.INVERSION, mode=terms.MODE_EVALUATE)
    expected = terms.TypedName(location=create_loc(1, 0, 1, 4), id='flag', ctx='load', mode=terms.MODE_EVALUATE)
    assert actual == expected


def test_comparison__single():
    actual = parse_expr('a', rn.COMPARISON, mode=terms.MODE_EVALUATE)
    expected = terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE)
    assert actual == expected


def test_comparison__chain():
    actual = parse_expr('a < b <= c', rn.COMPARISON, mode=terms.MODE_EVALUATE)
    expected = terms.Compare(
        location=create_loc(1, 0, 1, 10),
        left=terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE),
        operators=['<', '<='],
        comparators=[
            terms.TypedName(location=create_loc(1, 4, 1, 5), id='b', ctx='load', mode=terms.MODE_EVALUATE),
            terms.TypedName(location=create_loc(1, 9, 1, 10), id='c', ctx='load', mode=terms.MODE_EVALUATE),
        ],
    )
    assert actual == expected


def test_comparison__operators():
    operators = [
        '==',
        '!=',
        '<=',
        '<',
        '>=',
        '>',
        'not in',
        'in',
        'is not',
        'is',
    ]
    for op in operators:
        op_length = len(op)
        expr = f'a {op} b'
        actual = parse_expr(expr, rn.COMPARISON, mode=terms.MODE_EVALUATE)
        expected = terms.Compare(
            location=create_loc(1, 0, 1, 4 + op_length),
            left=terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE),
            operators=[op],
            comparators=[
                terms.TypedName(
                    location=create_loc(1, 3 + op_length, 1, 4 + op_length),
                    id='b',
                    ctx='load',
                    mode=terms.MODE_EVALUATE,
                ),
            ],
        )
        assert actual == expected, f'Failed for operator: {expr}'


def test_bitwise_or__chain():
    actual = parse_expr('a | b | c', rn.BITWISE_OR, mode=terms.MODE_EVALUATE)
    expected = terms.BinOp(
        location=create_loc(1, 0, 1, 9),
        left=terms.BinOp(
            location=create_loc(1, 0, 1, 5),
            left=terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE),
            op='|',
            right=terms.TypedName(location=create_loc(1, 4, 1, 5), id='b', ctx='load', mode=terms.MODE_EVALUATE),
        ),
        op='|',
        right=terms.TypedName(location=create_loc(1, 8, 1, 9), id='c', ctx='load', mode=terms.MODE_EVALUATE),
    )
    assert actual == expected


def test_bitwise_or__single():
    actual = parse_expr('a', rn.BITWISE_OR, mode=terms.MODE_EVALUATE)
    expected = terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE)
    assert actual == expected


def test_bitwise_xor__whitespace_required():
    actual = parse_expr('a ^b', rn.BITWISE_XOR, mode=terms.MODE_EVALUATE)
    assert isinstance(actual, syntax.ErrorTerm)


def test_bitwise_xor__chain():
    actual = parse_expr('a ^ b ^ c', rn.BITWISE_XOR, mode=terms.MODE_EVALUATE)
    expected = terms.BinOp(
        location=create_loc(1, 0, 1, 9),
        left=terms.BinOp(
            location=create_loc(1, 0, 1, 5),
            left=terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE),
            op='^',
            right=terms.TypedName(location=create_loc(1, 4, 1, 5), id='b', ctx='load', mode=terms.MODE_EVALUATE),
        ),
        op='^',
        right=terms.TypedName(location=create_loc(1, 8, 1, 9), id='c', ctx='load', mode=terms.MODE_EVALUATE),
    )
    assert actual == expected


def test_bitwise_xor__single():
    actual = parse_expr('a', rn.BITWISE_XOR, mode=terms.MODE_EVALUATE)
    expected = terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE)
    assert actual == expected


def test_bitwise_and__chain():
    actual = parse_expr('a & b & c', rn.BITWISE_AND, mode=terms.MODE_EVALUATE)
    expected = terms.BinOp(
        location=create_loc(1, 0, 1, 9),
        left=terms.BinOp(
            location=create_loc(1, 0, 1, 5),
            left=terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE),
            op='&',
            right=terms.TypedName(location=create_loc(1, 4, 1, 5), id='b', ctx='load', mode=terms.MODE_EVALUATE),
        ),
        op='&',
        right=terms.TypedName(location=create_loc(1, 8, 1, 9), id='c', ctx='load', mode=terms.MODE_EVALUATE),
    )
    assert actual == expected


def test_bitwise_and__single():
    actual = parse_expr('a', rn.BITWISE_AND, mode=terms.MODE_EVALUATE)
    expected = terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE)
    assert actual == expected


def test_bitwise_shift__chain():
    actual = parse_expr('a << b >> c', rn.SHIFT_EXPR, mode=terms.MODE_EVALUATE)
    expected = terms.BinOp(
        location=create_loc(1, 0, 1, 11),
        left=terms.BinOp(
            location=create_loc(1, 0, 1, 6),
            left=terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE),
            op='<<',
            right=terms.TypedName(location=create_loc(1, 5, 1, 6), id='b', ctx='load', mode=terms.MODE_EVALUATE),
        ),
        op='>>',
        right=terms.TypedName(location=create_loc(1, 10, 1, 11), id='c', ctx='load', mode=terms.MODE_EVALUATE),
    )
    assert actual == expected


def test_bitwise_shift__single():
    actual = parse_expr('a', rn.SHIFT_EXPR, mode=terms.MODE_EVALUATE)
    expected = terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE)
    assert actual == expected


def test_sum__invalid_arithmetic():
    actual = parse_expr('a + not b', rn.SUM, mode=terms.MODE_EVALUATE)
    expected = syntax.ErrorTerm(
        message="'not' after an operator must be parenthesized", location=create_loc(1, 0, 1, 9)
    )
    assert actual == expected


def test_sum__chain():
    actual = parse_expr('a + b - c', rn.SUM, mode=terms.MODE_EVALUATE)
    expected = terms.BinOp(
        location=create_loc(1, 0, 1, 9),
        left=terms.BinOp(
            location=create_loc(1, 0, 1, 5),
            left=terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE),
            op='+',
            right=terms.TypedName(location=create_loc(1, 4, 1, 5), id='b', ctx='load', mode=terms.MODE_EVALUATE),
        ),
        op='-',
        right=terms.TypedName(location=create_loc(1, 8, 1, 9), id='c', ctx='load', mode=terms.MODE_EVALUATE),
    )
    assert actual == expected


def test_sum__single():
    actual = parse_expr('a', rn.SUM, mode=terms.MODE_EVALUATE)
    expected = terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE)
    assert actual == expected


def test_term__chain():
    actual = parse_expr('a * b / c', rn.TERM, mode=terms.MODE_EVALUATE)
    expected = terms.BinOp(
        location=create_loc(1, 0, 1, 9),
        left=terms.BinOp(
            location=create_loc(1, 0, 1, 5),
            left=terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE),
            op='*',
            right=terms.TypedName(location=create_loc(1, 4, 1, 5), id='b', ctx='load', mode=terms.MODE_EVALUATE),
        ),
        op='/',
        right=terms.TypedName(location=create_loc(1, 8, 1, 9), id='c', ctx='load', mode=terms.MODE_EVALUATE),
    )
    assert actual == expected


def test_term__operators():
    operators = [
        '*',
        '/',
        '//',
        '%',
        '@',
    ]
    for op in operators:
        expr = f'a {op} b'
        actual = parse_expr(expr, rn.TERM, mode=terms.MODE_EVALUATE)
        expected = terms.BinOp(
            location=create_loc(1, 0, 1, 4 + len(op)),
            left=terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE),
            op=op,
            right=terms.TypedName(
                location=create_loc(1, 3 + len(op), 1, 4 + len(op)),
                id='b',
                ctx='load',
                mode=terms.MODE_EVALUATE,
            ),
        )
        assert actual == expected, f'Failed for operator: {expr}'


def test_term__single():
    actual = parse_expr('a', rn.TERM, mode=terms.MODE_EVALUATE)
    expected = terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE)
    assert actual == expected


def test_factor__invalid_factor():
    actual = parse_expr('+ not b', rn.FACTOR, mode=terms.MODE_EVALUATE)
    expected = syntax.ErrorTerm(
        message="'not' after an operator must be parenthesized", location=create_loc(1, 0, 1, 7)
    )
    assert actual == expected


def test_factor__operators():
    operators = [
        '+',
        '-',
        '~',
    ]
    for op in operators:
        expr = f'{op}b'
        actual = parse_expr(expr, rn.FACTOR, mode=terms.MODE_EVALUATE)
        expected = terms.UnaryOp(
            location=create_loc(1, 0, 1, 2 + len(op) - 1),
            op=op,
            operand=terms.TypedName(
                location=create_loc(1, 1 + len(op) - 1, 1, 2 + len(op) - 1),
                id='b',
                ctx='load',
                mode=terms.MODE_EVALUATE,
            ),
        )
        assert actual == expected, f'Failed for operator: {expr}'


def test_power__binary():
    actual = parse_expr('a ** b', rn.POWER, mode=terms.MODE_EVALUATE)
    expected = terms.BinOp(
        location=create_loc(1, 0, 1, 6),
        left=terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE),
        op='**',
        right=terms.TypedName(location=create_loc(1, 5, 1, 6), id='b', ctx='load', mode=terms.MODE_EVALUATE),
    )
    assert actual == expected


def test_power__single():
    actual = parse_expr('a', rn.POWER, mode=terms.MODE_EVALUATE)
    expected = terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE)
    assert actual == expected


def test_primary__attribute():
    actual = parse_expr('obj.attr', rn.PRIMARY, mode=terms.MODE_EVALUATE)
    expected = terms.Attribute(
        location=create_loc(1, 0, 1, 8),
        value=terms.TypedName(location=create_loc(1, 0, 1, 3), id='obj', ctx='load', mode=terms.MODE_EVALUATE),
        attr='attr',
        ctx='load',
    )
    assert actual == expected


def test_primary__call():
    actual = parse_expr('func(arg1, arg2)', rn.PRIMARY, mode=terms.MODE_EVALUATE)
    expected = terms.Call(
        location=create_loc(1, 0, 1, 16),
        func=terms.TypedName(location=create_loc(1, 0, 1, 4), id='func', ctx='load', mode=terms.MODE_EVALUATE),
        args=[
            terms.TypedName(location=create_loc(1, 5, 1, 9), id='arg1', ctx='load', mode=terms.MODE_EVALUATE),
            terms.TypedName(location=create_loc(1, 11, 1, 15), id='arg2', ctx='load', mode=terms.MODE_EVALUATE),
        ],
        keywords=[],
    )
    assert actual == expected


def test_primary__slice():
    actual = parse_expr('arr[10]', rn.PRIMARY, mode=terms.MODE_EVALUATE)
    expected = terms.Subscript(
        location=create_loc(1, 0, 1, 7),
        value=terms.TypedName(location=create_loc(1, 0, 1, 3), id='arr', ctx='load', mode=terms.MODE_EVALUATE),
        slice=terms.IntegerLiteral(value=10, mode=terms.MODE_EVALUATE, location=create_loc(1, 4, 1, 6)),
        ctx='load',
    )
    assert actual == expected


def test_primary__atom():
    actual = parse_expr('variable', rn.PRIMARY, mode=terms.MODE_EVALUATE)
    expected = terms.TypedName(location=create_loc(1, 0, 1, 8), id='variable', ctx='load', mode=terms.MODE_EVALUATE)
    assert actual == expected


def test_star_expressions__single():
    actual = parse_expr('a', rn.STAR_EXPRESSIONS, mode=terms.MODE_EVALUATE)
    expected = terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE)
    assert actual == expected


def test_star_expressions__multiple():
    actual = parse_expr('a, b, c', rn.STAR_EXPRESSIONS, mode=terms.MODE_EVALUATE)
    expected = terms.Tuple(
        location=create_loc(1, 0, 1, 7),
        elements=[
            terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE),
            terms.TypedName(location=create_loc(1, 3, 1, 4), id='b', ctx='load', mode=terms.MODE_EVALUATE),
            terms.TypedName(location=create_loc(1, 6, 1, 7), id='c', ctx='load', mode=terms.MODE_EVALUATE),
        ],
        ctx='load',
    )
    assert actual == expected


def test_star_named_epxressions__empty():
    actual = parse_expr(' ', rn.STAR_NAMED_EXPRESSIONS, mode=terms.MODE_EVALUATE)
    expected = syntax.ErrorTerm(message='chunk[line:1] Not all text consumed: indices 0:0/1:0.')
    assert actual == expected


def test_star_named_expressions__single():
    actual = parse_expr('a', rn.STAR_NAMED_EXPRESSIONS, mode=terms.MODE_EVALUATE)
    expected = syntax.TermList(
        terms=[terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE)],
    )
    assert actual == expected


def test_star_named_expressions__trailing_comma():
    actual = parse_expr('a,', rn.STAR_NAMED_EXPRESSIONS, mode=terms.MODE_EVALUATE)
    expected = syntax.TermList(
        terms=[terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE)],
    )
    assert actual == expected


def test_star_named_expressions__multiple():
    actual = parse_expr('a, b, c', rn.STAR_NAMED_EXPRESSIONS, mode=terms.MODE_EVALUATE)
    expected = syntax.TermList(
        terms=[
            terms.TypedName(location=create_loc(1, 0, 1, 1), id='a', ctx='load', mode=terms.MODE_EVALUATE),
            terms.TypedName(location=create_loc(1, 3, 1, 4), id='b', ctx='load', mode=terms.MODE_EVALUATE),
            terms.TypedName(location=create_loc(1, 6, 1, 7), id='c', ctx='load', mode=terms.MODE_EVALUATE),
        ],
    )
    assert actual == expected


def test_assignment_expression__simple():
    actual = parse_expr('x := 42', rn.ASSIGNMENT_EXPRESSION, mode=terms.MODE_EVALUATE)
    expected = terms.NamedExpr(
        location=create_loc(1, 0, 1, 7),
        target=terms.TypedName(location=create_loc(1, 0, 1, 1), id='x', ctx='store', mode=terms.MODE_EVALUATE),
        value=terms.IntegerLiteral(value=42, mode=terms.MODE_EVALUATE, location=create_loc(1, 5, 1, 7)),
    )
    assert actual == expected


def test_assignment_expression__expect_expression():
    actual = parse_expr('y := ', rn.ASSIGNMENT_EXPRESSION, mode=terms.MODE_EVALUATE)
    expected = syntax.ErrorTerm(message='Expected expression after ":="', location=create_loc(1, 0, 1, 4))
    assert actual == expected


def test_if_stmt__simple():
    actual = parse_expr('if x > 0:', rn.IF_STMT, mode=terms.MODE_EVALUATE)
    expected = terms.TypedIf(
        location=create_loc(1, 0, 1, 9),
        test=terms.Compare(
            location=create_loc(1, 2, 1, 8),
            left=terms.TypedName(location=create_loc(1, 3, 1, 4), id='x', ctx='load', mode=terms.MODE_EVALUATE),
            operators=['>'],
            comparators=[terms.IntegerLiteral(value=0, mode=terms.MODE_EVALUATE, location=create_loc(1, 7, 1, 8))],
        ),
        body=syntax.TermList(terms=[], is_placeholder=True),
        elifs=[],
        orelse=syntax.Empty,
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_if_stmt__named():
    actual = parse_expr('if (a := 42):', rn.IF_STMT, mode=terms.MODE_EVALUATE)
    expected = terms.TypedIf(
        location=create_loc(1, 0, 1, 13),
        test=terms.NamedExpr(
            location=create_loc(1, 4, 1, 11),
            target=terms.TypedName(location=create_loc(1, 4, 1, 5), id='a', ctx='store', mode=terms.MODE_EVALUATE),
            value=terms.IntegerLiteral(value=42, mode=terms.MODE_EVALUATE, location=create_loc(1, 9, 1, 11)),
        ),
        body=syntax.TermList(terms=[], is_placeholder=True),
        elifs=[],
        orelse=syntax.Empty,
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_elif_stmt__simple():
    actual = parse_expr('elif y < 10:', rn.ELIF_STMT, mode=terms.MODE_EVALUATE)
    expected = terms.ElifSibling(
        location=create_loc(1, 0, 1, 12),
        test=terms.Compare(
            location=create_loc(1, 4, 1, 11),
            left=terms.TypedName(location=create_loc(1, 5, 1, 6), id='y', ctx='load', mode=terms.MODE_EVALUATE),
            operators=['<'],
            comparators=[terms.IntegerLiteral(value=10, mode=terms.MODE_EVALUATE, location=create_loc(1, 9, 1, 11))],
        ),
        body=syntax.TermList(terms=[], is_placeholder=True),
    )
    assert actual == expected


def test_else_stmt__simple():
    actual = parse_expr('else:', rn.ELSE_BLOCK, mode=terms.MODE_EVALUATE)
    expected = terms.ElseSibling(
        location=create_loc(1, 0, 1, 5),
        body=syntax.TermList(terms=[], is_placeholder=True),
    )
    assert actual == expected


def test_for_stmt__simple():
    actual = parse_expr('for item in collection:', rn.FOR_STMT, mode=terms.MODE_EVALUATE)
    expected = terms.TypedFor(
        location=create_loc(1, 0, 1, 23),
        target=terms.TypedName(location=create_loc(1, 4, 1, 8), id='item', ctx='store', mode=terms.MODE_EVALUATE),
        iter=terms.TypedName(location=create_loc(1, 12, 1, 22), id='collection', ctx='load', mode=terms.MODE_EVALUATE),
        body=syntax.TermList(terms=[], is_placeholder=True),
        orelse=syntax.Empty,
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_for_stmt__tuple_target():
    actual = parse_expr('for a, b in pairs:', rn.FOR_STMT, mode=terms.MODE_EVALUATE)
    expected = terms.TypedFor(
        location=create_loc(1, 0, 1, 18),
        target=terms.Tuple(
            location=create_loc(1, 3, 1, 8),
            elements=[
                terms.TypedName(location=create_loc(1, 4, 1, 5), id='a', ctx='store', mode=terms.MODE_EVALUATE),
                terms.TypedName(location=create_loc(1, 7, 1, 8), id='b', ctx='store', mode=terms.MODE_EVALUATE),
            ],
            ctx='store',
        ),
        iter=terms.TypedName(location=create_loc(1, 12, 1, 17), id='pairs', ctx='load', mode=terms.MODE_EVALUATE),
        body=syntax.TermList(terms=[], is_placeholder=True),
        orelse=syntax.Empty,
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_while_stmt__simple():
    actual = parse_expr('while condition:', rn.WHILE_STMT, mode=terms.MODE_EVALUATE)
    expected = terms.TypedWhile(
        location=create_loc(1, 0, 1, 16),
        test=terms.TypedName(location=create_loc(1, 6, 1, 15), id='condition', ctx='load', mode=terms.MODE_EVALUATE),
        body=syntax.TermList(terms=[], is_placeholder=True),
        orelse=syntax.Empty,
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_while_stmt__named():
    actual = parse_expr('while x := 42:', rn.WHILE_STMT, mode=terms.MODE_EVALUATE)
    expected = terms.TypedWhile(
        location=create_loc(1, 0, 1, 14),
        test=terms.NamedExpr(
            location=create_loc(1, 5, 1, 13),
            target=terms.TypedName(location=create_loc(1, 6, 1, 7), id='x', ctx='store', mode=terms.MODE_EVALUATE),
            value=terms.IntegerLiteral(value=42, mode=terms.MODE_EVALUATE, location=create_loc(1, 11, 1, 13)),
        ),
        body=syntax.TermList(terms=[], is_placeholder=True),
        orelse=syntax.Empty,
        mode=terms.MODE_EVALUATE,
    )
    assert actual == expected


def test_with_stmt__no_as():
    actual = parse_expr('with resource:', rn.WITH_STMT, mode=terms.MODE_EVALUATE)
    expected = terms.With(
        location=create_loc(1, 0, 1, 14),
        items=[
            terms.WithItem(
                context_expr=terms.TypedName(
                    location=create_loc(1, 5, 1, 13), id='resource', ctx='load', mode=terms.MODE_EVALUATE
                ),
                optional_vars=syntax.Empty,
            ),
        ],
        body=syntax.TermList(terms=[], is_placeholder=True),
    )
    assert actual == expected


def test_with_stmt__simple():
    actual = parse_expr('with resource as res:', rn.WITH_STMT, mode=terms.MODE_EVALUATE)
    expected = terms.With(
        location=create_loc(1, 0, 1, 21),
        items=[
            terms.WithItem(
                context_expr=terms.TypedName(
                    location=create_loc(1, 5, 1, 13), id='resource', ctx='load', mode=terms.MODE_EVALUATE
                ),
                optional_vars=terms.TypedName(
                    location=create_loc(1, 17, 1, 20), id='res', ctx='store', mode=terms.MODE_EVALUATE
                ),
            ),
        ],
        body=syntax.TermList(terms=[], is_placeholder=True),
    )
    assert actual == expected


def test_with_stmt__multiple_items():
    actual = parse_expr('with res1, res2 as r2:', rn.WITH_STMT, mode=terms.MODE_EVALUATE)
    expected = terms.With(
        location=create_loc(1, 0, 1, 22),
        items=[
            terms.WithItem(
                context_expr=terms.TypedName(
                    location=create_loc(1, 5, 1, 9), id='res1', ctx='load', mode=terms.MODE_EVALUATE
                ),
                optional_vars=syntax.Empty,
            ),
            terms.WithItem(
                context_expr=terms.TypedName(
                    location=create_loc(1, 11, 1, 15), id='res2', ctx='load', mode=terms.MODE_EVALUATE
                ),
                optional_vars=terms.TypedName(
                    location=create_loc(1, 19, 1, 21), id='r2', ctx='store', mode=terms.MODE_EVALUATE
                ),
            ),
        ],
        body=syntax.TermList(terms=[], is_placeholder=True),
    )
    assert actual == expected


def test_expression__double_layer():
    actual = parse_expr('<3:A>', rn.EXPRESSION)
    expected = syntax.Layers(
        layers=[
            terms.IntegerLiteral(value=3, mode=terms.MODE_EVALUATE, location=create_loc(1, 1, 1, 2)),
            terms.TypedName(id='A', ctx='load', mode=terms.MODE_TYPECHECK, location=create_loc(1, 3, 1, 4)),
        ],
    )
    assert actual == expected


def test_expression__name_bang():
    actual = parse_expr('Dog!', rn.EXPRESSION, mode=terms.MODE_EVALUATE)
    expected = terms.Attribute(
        value=terms.TypedName(id='Dog', ctx='load', mode=terms.MODE_EVALUATE, location=create_loc(1, 0, 1, 3)),
        attr='result__sa',
        ctx='load',
        location=create_loc(1, 0, 1, 4),
    )
    assert actual == expected


def test_expression__bang():
    actual = parse_expr('Matrix(3,5)!', rn.EXPRESSION, mode=terms.MODE_EVALUATE)
    expected = terms.Attribute(
        value=terms.Call(
            func=terms.TypedName(id='Matrix', ctx='load', mode=terms.MODE_EVALUATE, location=create_loc(1, 0, 1, 6)),
            args=[
                terms.IntegerLiteral(value=3, mode=terms.MODE_EVALUATE, location=create_loc(1, 7, 1, 8)),
                terms.IntegerLiteral(value=5, mode=terms.MODE_EVALUATE, location=create_loc(1, 9, 1, 10)),
            ],
            keywords=[],
            location=create_loc(1, 0, 1, 11),
        ),
        attr='result__sa',
        ctx='load',
        location=create_loc(1, 0, 1, 12),
    )
    assert actual == expected



from tapl_lang.lib import tapl_dev

class SimplestClass:
    pass

def accept(param):
    pass
accept(SimplestClass())

class Circle:

    def __init__(self, radius):
        self.radius = radius

    def area(self):
        return 3.14 * self.radius * self.radius

    def set_radius(self, radius):
        self.radius = radius

    def is_bigger_than(self, radius):
        return self.radius > radius

def print_area(circle):
    tapl_dev.log(circle.area())
    tapl_dev.log(circle.is_bigger_than(1.0))
print_area(Circle(2.0))



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import ast

from tapl_lang.core import syntax
from tapl_lang.lib import python_backend, terms


def test_generate_expr():
    backend = python_backend.AstGenerator()
    setting = syntax.BackendSetting(scope_level=0)
    term = terms.TypedName(
        location=syntax.Location(start=syntax.Position(line=1, column=0), end=syntax.Position(line=1, column=1)),
        id='x',
        ctx='load',
        mode=terms.MODE_EVALUATE,
    )
    expr = backend.try_generate_expr(term, setting)
    assert expr is None
    expr = backend.generate_expr(term, setting)
    assert isinstance(expr, ast.expr)
    assert isinstance(expr, ast.Name)
    assert expr.id == 'x'
    assert isinstance(expr.ctx, ast.Load)


def test_generate_stmt():
    backend = python_backend.AstGenerator()
    setting = syntax.BackendSetting(scope_level=0)
    term = terms.TypedReturn(
        location=syntax.Location(start=syntax.Position(line=1, column=0), end=syntax.Position(line=1, column=1)),
        value=terms.Constant(
            location=syntax.Location(start=syntax.Position(line=1, column=7), end=syntax.Position(line=1, column=9)),
            value=42,
        ),
        mode=terms.MODE_EVALUATE,
    )
    stmts = backend.try_generate_stmt(term, setting)
    assert stmts is None
    stmts = backend.generate_stmt(term, setting)
    assert isinstance(stmts, list)
    assert len(stmts) == 1
    stmt = stmts[0]
    assert isinstance(stmt, ast.stmt)
    assert isinstance(stmt, ast.Return)



from tapl_lang.lib import tapl_typing
from tapl_lang.pythonlike.predef1 import predef_scope as predef_scope__sa
s0 = tapl_typing.create_scope(parent__sa=predef_scope__sa)

def Matrix(rows, cols):
    s1 = tapl_typing.create_scope(parent__sa=s0, rows=rows, cols=cols)

    class Matrix_:
        class_name = 'Matrix({},{})'.format(s1.rows, s1.cols)

        def __init__(self):
            s2 = tapl_typing.create_scope(parent__sa=s1, self=self)
            s2.self.rows = s2.rows
            s2.self.cols = s2.cols
            s2.self.num_rows = s2.Int
            s2.self.num_cols = s2.Int
            s2.self.values = s2.List(s2.List(s2.Int))
            with tapl_typing.scope_forker(s2) as f2:
                s3 = tapl_typing.fork_scope(f2)
                s3.i = s3.range(s3.self.num_rows).__iter__().__next__()
                s3.columns = s3.List(s3.Int)
                with tapl_typing.scope_forker(s3) as f3:
                    s4 = tapl_typing.fork_scope(f3)
                    s4.j = s4.range(s4.self.num_cols).__iter__().__next__()
                    s4.columns.append(s4.Int)
                    s4 = tapl_typing.fork_scope(f3)
                s3.self.values.append(s3.columns)
                s3 = tapl_typing.fork_scope(f2)
            return tapl_typing.get_return_type(s2)

        def __repr__(self):
            s2 = tapl_typing.create_scope(parent__sa=s1, self=self)
            tapl_typing.add_return_type(s2, s2.str(s2.self.values))
            return tapl_typing.get_return_type(s2)
    s1.Matrix_ = tapl_typing.create_class(cls=Matrix_, init_args=[], methods=[('__repr__', [])])
    tapl_typing.add_return_type(s1, s1.Matrix_)
    return tapl_typing.get_return_type(s1)
s0.Matrix = Matrix

def accept_matrix_2_3(matrix):
    s1 = tapl_typing.create_scope(parent__sa=s0, matrix=matrix)
    pass
    return tapl_typing.get_return_type(s1)
s0.accept_matrix_2_3 = tapl_typing.create_function([s0.Matrix(2, 3).result__sa], accept_matrix_2_3(s0.Matrix(2, 3).result__sa))

def add(rows, cols):
    s1 = tapl_typing.create_scope(parent__sa=s0, rows=rows, cols=cols)

    def add_(a, b):
        s2 = tapl_typing.create_scope(parent__sa=s1, a=a, b=b)
        s2.result = s2.Matrix(s2.rows, s2.cols)()
        with tapl_typing.scope_forker(s2) as f2:
            s3 = tapl_typing.fork_scope(f2)
            s3.i = s3.range(s3.result.num_rows).__iter__().__next__()
            with tapl_typing.scope_forker(s3) as f3:
                s4 = tapl_typing.fork_scope(f3)
                s4.j = s4.range(s4.result.num_cols).__iter__().__next__()
                s4.result.values[s4.i][s4.j] = s4.a.values[s4.i][s4.j] + s4.b.values[s4.i][s4.j]
                s4 = tapl_typing.fork_scope(f3)
            s3 = tapl_typing.fork_scope(f2)
        tapl_typing.add_return_type(s2, s2.result)
        return tapl_typing.get_return_type(s2)
    s1.add_ = tapl_typing.create_function([s1.Matrix(s1.rows, s1.cols).result__sa, s1.Matrix(s1.rows, s1.cols).result__sa], add_(s1.Matrix(s1.rows, s1.cols).result__sa, s1.Matrix(s1.rows, s1.cols).result__sa))
    tapl_typing.add_return_type(s1, s1.add_)
    return tapl_typing.get_return_type(s1)
s0.add = add

def multiply(m, n, p):
    s1 = tapl_typing.create_scope(parent__sa=s0, m=m, n=n, p=p)

    def multiply_(a, b):
        s2 = tapl_typing.create_scope(parent__sa=s1, a=a, b=b)
        s2.result = s2.Matrix(s2.m, s2.p)()
        with tapl_typing.scope_forker(s2) as f2:
            s3 = tapl_typing.fork_scope(f2)
            s3.i = s3.range(s3.a.num_rows).__iter__().__next__()
            with tapl_typing.scope_forker(s3) as f3:
                s4 = tapl_typing.fork_scope(f3)
                s4.j = s4.range(s4.b.num_cols).__iter__().__next__()
                with tapl_typing.scope_forker(s4) as f4:
                    s5 = tapl_typing.fork_scope(f4)
                    s5.k = s5.range(s5.a.num_cols).__iter__().__next__()
                    s5.result.values[s5.i][s5.j] = s5.result.values[s5.i][s5.j] + s5.a.values[s5.i][s5.k] * s5.b.values[s5.k][s5.j]
                    s5 = tapl_typing.fork_scope(f4)
                s4 = tapl_typing.fork_scope(f3)
            s3 = tapl_typing.fork_scope(f2)
        tapl_typing.add_return_type(s2, s2.result)
        return tapl_typing.get_return_type(s2)
    s1.multiply_ = tapl_typing.create_function([s1.Matrix(s1.m, s1.n).result__sa, s1.Matrix(s1.n, s1.p).result__sa], multiply_(s1.Matrix(s1.m, s1.n).result__sa, s1.Matrix(s1.n, s1.p).result__sa))
    tapl_typing.add_return_type(s1, s1.multiply_)
    return tapl_typing.get_return_type(s1)
s0.multiply = multiply

def main():
    s1 = tapl_typing.create_scope(parent__sa=s0)
    s1.matrix_2_2 = s1.Matrix(2, 2)()
    s1.matrix_2_2.values = tapl_typing.create_typed_list(tapl_typing.create_typed_list(s1.Int, s1.Int), tapl_typing.create_typed_list(s1.Int, s1.Int))
    s1.matrix_2_3 = s1.Matrix(2, 3)()
    s1.matrix_2_3.values = tapl_typing.create_typed_list(tapl_typing.create_typed_list(s1.Int, s1.Int, s1.Int), tapl_typing.create_typed_list(s1.Int, s1.Int, s1.Int))
    s1.accept_matrix_2_3(s1.matrix_2_3)
    s1.print(s1.add(2, 2)(s1.matrix_2_2, s1.matrix_2_2))
    s1.print(s1.multiply(2, 2, 3)(s1.matrix_2_2, s1.matrix_2_3))
    return tapl_typing.get_return_type(s1)
s0.main = tapl_typing.create_function([], main())
with tapl_typing.scope_forker(s0) as f0:
    s1 = tapl_typing.fork_scope(f0)
    s1.__name__ == s1.Str
    s1.main()
    s1 = tapl_typing.fork_scope(f0)


# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from tapl_lang.core import line_record, tapl_error


class Chunk:
    def __init__(self, line_records: list[line_record.LineRecord], children: list['Chunk']) -> None:
        self.line_records = line_records
        self.children = children


def min_indentation(line_records: list[line_record.LineRecord]) -> int | None:
    min_indent = None
    for line in line_records:
        if line.empty:
            continue
        if min_indent is None:
            min_indent = line.indent
        elif line.indent is not None:
            min_indent = min(min_indent, line.indent)
    return min_indent


def get_same_indent_indexes(line_records: list[line_record.LineRecord], indent: int) -> list[int]:
    result = [i for i in range(len(line_records)) if line_records[i].indent == indent]
    first_non_empty = next((i for i in range(len(line_records)) if not line_records[i].empty), None)
    if len(result) > 0 and result[0] != first_non_empty:
        raise tapl_error.TaplError('First line should have the same indent as the given indent.')
    return result


def find_first_ends_with_colon(line_records: list[line_record.LineRecord]) -> int | None:
    for i in range(len(line_records)):
        if line_records[i].ends_with_colon:
            return i
    return None


class Chunker:
    def decode_chunk(self, line_records: list[line_record.LineRecord]) -> Chunk:
        index = find_first_ends_with_colon(line_records)
        if index is not None:
            children_start_index = index + 1
            children = self.decode_chunks(line_records[children_start_index:])
            return Chunk(line_records[:children_start_index], children)
        return Chunk(line_records, [])

    def decode_chunks(self, line_records: list[line_record.LineRecord]) -> list[Chunk]:
        min_indent = min_indentation(line_records)
        # If all line records are empty
        if min_indent is None:
            return []
        indexes = get_same_indent_indexes(line_records, min_indent)
        chunks: list[Chunk] = []
        for i in range(len(indexes)):
            start_index = indexes[i]
            if i + 1 < len(indexes):
                end_index = indexes[i + 1]
            else:
                end_index = len(line_records)
            chunk = self.decode_chunk(line_records[start_index:end_index])
            chunks.append(chunk)
        return chunks


def chunk_text(text: str) -> list[Chunk]:
    return Chunker().decode_chunks(line_record.split_text_to_lines(text))



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from typing import Any

from tapl_lang.lib import builtin_types as bt
from tapl_lang.lib import kinds

python_builtin_types = {
    '__name__': bt.Str,
    'print': kinds.create_function([bt.Any], bt.NoneType),
    'range': kinds.create_function([bt.Int], [bt.Int]),
    'str': kinds.create_function([bt.Any], bt.Str),
    'abs': kinds.create_function([bt.Float], bt.Float),
    'round': kinds.create_function([bt.Float], bt.Float),
}

export1: dict[str, Any] = {}
export1.update(python_builtin_types)
export1.update(bt.Types)
export1.update(bt.TYPE_CONSTRUCTORS)



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

import pytest

from tapl_lang.lib import builtin_types, scope


def test_variable_from_parent_scope():
    parent_scope = scope.Scope()
    parent_scope.y = 'parent_value'
    child_scope = scope.Scope(parent=parent_scope)
    assert child_scope.y == 'parent_value'


def test_reassign_variable():
    s = scope.Scope()
    s.store__sa('x', builtin_types.Int)
    assert s.load__sa('x') == builtin_types.Int
    s.store__sa('x', builtin_types.Int)
    assert s.load__sa('x') == builtin_types.Int
    with pytest.raises(TypeError, match=r'Type error in variable "x": Expected type "Int", but found "Str"\.'):
        s.store__sa('x', builtin_types.Str)



from tapl_lang.lib import tapl_typing
from tapl_lang.pythonlike.predef1 import predef_scope as predef_scope__sa
s0 = tapl_typing.create_scope(parent__sa=predef_scope__sa)

def one():
    s1 = tapl_typing.create_scope(parent__sa=s0)
    s1.one = tapl_typing.create_function([], s1.Str)
    tapl_typing.set_return_type(s1, s1.Str)
    tapl_typing.add_return_type(s1, s1.Int)
    return tapl_typing.get_return_type(s1)
s0.one = tapl_typing.create_function([], one())



from tapl_lang.lib import tapl_dev

def collatz_sequence(n):
    sequence = []
    if n < 1:
        return sequence
    while n != 1:
        if n % 2 == 0:
            n = n // 2
        else:
            n = 3 * n + 1
        sequence.append(n)
    return sequence
tapl_dev.log(tapl_dev.describe(collatz_sequence))
print(collatz_sequence(7))
print(collatz_sequence(19))
print(collatz_sequence(0))
tapl_dev.log(collatz_sequence(4))



# Part of the Tapl Language project, under the Apache License v2.0 with LLVM
# Exceptions. See /LICENSE for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from __future__ import annotations

from typing import Any

from typing_extensions import Self

from tapl_lang.lib import dynamic_attribute, kinds


class Slot:
    def __init__(self, value: Any):
        self.value = value


class Scope(dynamic_attribute.DynamicAttributeMixin):
    def __init__(self, parent: Scope | None = None, fields: dict[str, Any] | None = None, label__sa: str | None = None):
        self.parent__sa = parent
        self.fields__sa: dict[str, Slot] = {}
        if fields:
            self.store_many__sa(fields)
        self.return_type__sa = None
        self.returns__sa: list[Any] = []
        self.label__sa = label__sa

    def find_slot__sa(self, name: str) -> Slot | None:
        if name in self.fields__sa:
            return self.fields__sa[name]
        if self.parent__sa is not None:
            return self.parent__sa.find_slot__sa(name)
        return None

    def load__sa(self, name: str) -> Any:
        slot = self.find_slot__sa(name)
        if slot is not None:
            return slot.value
        raise AttributeError(f'{self.get_label__sa()} scope has no attribute "{name}"')

    def store__sa(self, name: str, value: Any) -> None:
        slot = self.find_slot__sa(name)
        if slot is None:
            self.fields__sa[name] = Slot(value)
            return
        if not kinds.check_subtype(value, slot.value):
            raise TypeError(f'Type error in variable "{name}": Expected type "{slot.value}", but found "{value}".')

    def store_many__sa(self, fields: dict[str, Any]) -> None:
        for name, value in fields.items():
            self.store__sa(name, value)

    def get_label__sa(self) -> str:
        if self.label__sa:
            return self.label__sa
        if self.parent__sa is None:
            return 'Global'
        return f'Scope(parent={self.parent__sa.get_label__sa()})'

    def __repr__(self) -> str:
        if '__repr__' in self.fields__sa:
            return self.fields__sa['__repr__'].value()
        if self.label__sa:
            return self.label__sa
        if self.parent__sa is None:
            return f'Scope(parent={self.parent__sa})'
        return object.__repr__(self)


class ScopeForker:
    def __init__(self, scope: Scope):
        self.parent = scope
        self.branches: list[Scope] = []

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        if not self.branches:
            raise RuntimeError('No branches were created in the scope forker.')
        for var in self.branches[0].fields__sa:
            values = []
            for record in self.branches:
                slot = record.fields__sa.get(var)
                if slot is not None:
                    values.append(slot.value)
            if len(values) == len(self.branches):
                self.parent.store__sa(var, kinds.create_union(*values))

    def new_scope(self) -> Scope:
        forked = Scope(parent=self.parent)
        self.branches.append(forked)
        return forked
